package dacia;

import java.awt.*;
import java.awt.event.*;
import java.awt.Font;

public class Chat extends Transform {
  ChatFrame frame = null;
  String text = null;

  public Chat() {
    super("Chat");
    frame = new ChatFrame(this);
    frame.init();
  }

  public void handleMessage(Message msg) {
    frame.displayMessage((String)msg.getData());
  }

  void sendMessage(String message) {
    Message msg = new Message();
    msg.setData((Object)message);
    output(1, msg, 1);
  }

  public void pack() {
    text = frame.getText();
    frame.quit();
    frame = null;
  }

  public void unpack() {
    frame = new ChatFrame(this);
    frame.init();
    frame.displayText(text);
  }

  public static void main(String[] args) {
    Chat chat = new Chat();
  }
}


class ChatFrame implements ActionListener{
  Chat chat;
  Frame f = null;
  Button sendButton = null;
  MenuItem startItem = null;
  MenuItem stopItem = null;
  TextField editText = null;
  TextArea displayArea = null;

  public ChatFrame(Chat target) {
    chat = target;
  }

  void init() {
    f = new Frame("Chat");
    f.setSize(350, 250);
    Font font = new Font("sanserif", Font.PLAIN|Font.BOLD, 14);
    f.setFont(font);
    f.show();

    // create the user interface
    GridBagLayout gridbag = new GridBagLayout();
    f.setLayout(gridbag);

    sendButton = new Button("Send");
    editText = new TextField();
    displayArea = new TextArea(7,40);
    displayArea.setEditable(false);

    this.constrain(f, new Label("Enter a new message:"), 0, 0, 2, 1, 0, 0, 0, 0);
    this.constrain(f, editText, 0, 1, 1, 1, 
		   GridBagConstraints.HORIZONTAL,
		   GridBagConstraints.NORTHWEST,
		   10.0, 0.0, 0, 0, 0, 0);
    this.constrain(f, sendButton, 1, 1, 1, 1, 
		   GridBagConstraints.HORIZONTAL,
		   GridBagConstraints.NORTHWEST,
		   1, 0.0, 0, 0, 0, 0);
    this.constrain(f, new Label("Latest message:"), 0, 2, 2, 1, 
		   GridBagConstraints.HORIZONTAL,
		   GridBagConstraints.NORTHWEST,
		   1.0, 0.0, 0, 0, 0, 0);
    this.constrain(f, displayArea, 0, 3, 2, 1, 
		   GridBagConstraints.BOTH,
		   GridBagConstraints.CENTER,
		   1.0, 1.0, 0, 0, 1, 0);

    // set up the menu bar
    Menu aMenu = new Menu("Menu");
    startItem = new MenuItem("Start");
    startItem.addActionListener((ActionListener)this);
    aMenu.add(startItem);
    stopItem = new MenuItem("Quit");
    stopItem.addActionListener((ActionListener)this);
    aMenu.add(stopItem);
    MenuBar aMenuBar = new MenuBar();
    aMenuBar.add(aMenu);
    f.setMenuBar(aMenuBar);
    sendButton.addActionListener(this);

		// Handle window close requests
		f.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
        f.removeAll();
				f.dispose(); 
      }
		});  
}

  //process menu commands
  public void actionPerformed(ActionEvent e) {
    try{
      if (e.getActionCommand().equals("Start")) {
      }
      else if (e.getActionCommand().equals("Quit")) {
        f.setVisible(false);
        f.dispose();
        //System.exit(0);
        //Thread.currentThread().stop();
      }
      else if (e.getActionCommand().equals("Send")) {
        displayArea.append("\nOUT: " + editText.getText());
        chat.sendMessage(editText.getText());
        editText.setText("");
        editText.requestFocus();
      }
      else
        System.out.println("Chat: unknown menu item: " + e.getActionCommand());
    }
    catch(Exception ex) {
      System.out.println("Chat: exception: "+ex.getMessage());
      ex.printStackTrace();
    }
  }

  void quit() {
    f.setVisible(false);
    f.dispose();
  }

  // the gridbag layout stuff from the Java NutShell book
  void constrain(Container container, Component component,
			int grid_x, int grid_y, int grid_width, int grid_height,
			int fill, int anchor, double weight_x, double weight_y,
			int top, int left, int bottom, int right) {
    GridBagConstraints c = new GridBagConstraints();
    c.gridx = grid_x; c.gridy = grid_y;
    c.gridwidth = grid_width; c.gridheight = grid_height;
    c.fill = fill; c.anchor = anchor;
    c.weightx = weight_x; c.weighty = weight_y;
    if (top+bottom+left+right > 0)
      c.insets = new Insets(top, left, bottom, right);
    
    ((GridBagLayout)container.getLayout()).setConstraints(component, c);
    container.add(component);
  }

  void constrain(Container container, Component component,
			int grid_x, int grid_y, int grid_width, int grid_height,
			int top, int left, int bottom, int right) {
    constrain(container, component, grid_x, grid_y,
              grid_width, grid_height, GridBagConstraints.NONE,
              GridBagConstraints.NORTHWEST,
              0.0, 0.0, top, left, bottom, right);
  }

  String getText() {
    return displayArea.getText();
  }

  void displayText(String str) {
    displayArea.append(str);
  }

  void displayMessage(String str) {
    displayArea.append("\nIN : " + str);
  }
}


