/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 2000 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: radu@eecs.umich.edu, caoy@umich.edu
*******************************************************************/

package dacia; 

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * DClassLoader is extended from ClassLoader. It first check the local disk for the 
 * requested class. If it is not find locally, it sends the request to a host connected 
 * with the Engine. The class be received from the remote host if it is available.  
 * @see Engine
 * @author Yizhou Cao
 */
public class DClassLoader extends ClassLoader {

  private String root;
  private Connection connection;

  public DClassLoader () {
      connection = null;
  }
  
/**
  * Set the connection.
  */
 public void setConnection (Connection connection) {
      this.connection = connection;
  }

/**
  * Load the class from network and define the class.
  */
  public Class findClass(String name) {
	  System.out.println("findlass(): finding class: " + name);
	  byte[] b = loadClassFromNetwork(name);
	  if (b == null || b.length == 0)
		  return null;
    
    return defineClass(name, b, 0, b.length);
  }

  private byte[] loadClassFromNetwork(String name) {
    Engine.classBuffer = null;

      // Create a message requesting for the class
	  NetworkMessage netmsg = new NetworkMessage(Engine.LOAD);
	  netmsg.data = (Object)name;

      // Send the request
    if (connection != null) {
 		  netmsg.sendMessage(connection);		
		  synchronized(Thread.currentThread()) {
        try {
            Thread.currentThread().sleep(1000);
        }
        catch(InterruptedException e){
            System.out.println("loadClassFromNetwork(): exception in wait(): " + e.getMessage());
            e.printStackTrace();
        }
      }	 
    }
    else { // Find a live connection
 	    Connection conn;
	    int nconns = Engine.connections.size();
	
      if (nconns == 0) {
        System.out.println("loadClassFromNetwork(): No remote Engine available");
        return null;
      }
	    for (int i=0; i<nconns; i++) {	  
	// Send the request to the connections one by one until a host returns the class
        conn = (Connection)Engine.connections.elementAt(i);
        System.out.println("loadClassFromNetwork(): send request to connection "
                           + i + " for " + name);
        netmsg.sendMessage(conn);		
        synchronized(Thread.currentThread()) {
          try {
            Thread.currentThread().sleep(1000);	// Wait for the class to be returned
          }
          catch(InterruptedException e) {
            System.out.println("loadClassFromNetwork(): exception in wait(): " + e.getMessage());
            e.printStackTrace();
          }
        }		  
        if (Engine.classBuffer != null) 
            break;
      }
    }
     
	// If the class is returned, save it to the local file system.
    if (Engine.classBuffer != null) {
        String filename = name.replace ('.', File.separatorChar) + ".class";
        File f = new File (filename); 
        try {
            FileOutputStream fos = new FileOutputStream(f);
            fos.write(Engine.classBuffer);
            fos.close();
        }
        catch(Exception e){
            System.out.println("loadClassFromNetwork():  " + e.getMessage());
        }
        return Engine.classBuffer;
    }
	
    return null;
  }
  
}
