package dacia;

/* Implements the FILTER PROC, having 1 input (port 0) and 1 output (port 1)
 * ports. Filters the messages received on the input port and outputs only
 * the messages matching the filter. It does not modify messages.
 */
public abstract class Filter extends Proc {
  public Filter(String procName) {
		super(procName, 2);
		type = FILTER;
	}

	public void handleMessage(Message msg, int port) {
    if(port == 1)
      System.out.println("Filter " + id + " : message received on port 1");
		else if(filterMatched(msg))
		  output(1, msg, 1);
	}

	public void handleAsyncMessage() {
		Message msg = null;
    while(true) {
      if((msg = getMessage()) == null)
        return;
      if(msg.getPort() == 1)
        System.out.println("Filter " + id +
                           " : async message received on port 1");
      else if(filterMatched(msg)) {
        output(1, msg, 0);
      }
    }
	}

  /* This abstract method should be defined in the subclass.
	 * It returns true if the message matches the filter.
   */
	public abstract boolean filterMatched(Message msg);
}
