/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 2000 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: radu@eecs.umich.edu, caoy@umich.edu
*******************************************************************/

package dacia;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.net.*;


/**
 * A box is used to represent each host.
 */
class Host {  
    /** name of the host */
    String name;  
	/** x coordinate */
    int x;  
	/** y coordinate */
	int y;  
    /** width in pixels */
	int w;   
    /** height */
	int h;   
	/** index of the color */
	int color;  
} 
  
/**
 * A cable represents the connection between hosts.
 */
class Cable {     
    /** name of the host connected from */
    String from;
    /** name of the host connected to */
    String to;
}  

/**
 * The small rectangular box, called node, represents each PROC.
 */
class Node {
	/** x coordinate */
    int x;
	/** y coordinate */
    int y;

    /** width */
    int w;
    /** height */
    int h;

    double dx;
    double dy;

    /** indicating if the PROC is local */
    boolean local; 	 
    /** PROC ID */	
    int	   id;
    /** the host name */
    String host; 	
    /** the PROC's name */
    String name;
}

/**
 * A edge represents the connection between PROCs.
 */
class Edge {
	/** the ID of the PROC it is connected from */
    int from;
	/** the ID of the PROC it is connected to */
    int to;

    double len;
}

/**
 * This is the popup menu.
 */
class Popup extends PopupMenu {
    final static int POPUP_PROC = 0; 
    final static int POPUP_ENGINE = 1; 	
    
    Popup(ActionListener listener, int mode) {
		super();
		MenuItem m;

		// menu for PROC operations
		m = new MenuItem("Start Proc");
		m.addActionListener(listener); 	
		add(m); 
		m = new MenuItem("Connect Procs");
		m.addActionListener(listener); 	
		add(m); 
		m = new MenuItem("Disconnect Proc");
		m.addActionListener(listener); 
		add(m); 	
		m = new MenuItem("Move Proc");
		m.addActionListener(listener); 	
		add(m);  	
		m = new MenuItem("Remove Proc");
		m.addActionListener(listener); 	
		add(m);  	
		if (mode == POPUP_PROC) 
			return; 		
		
		// menu for Engine operations
		addSeparator();
		m = new MenuItem("Connect Engines");
		m.addActionListener(listener); 
		add(m); 
		m = new MenuItem("Start Monitor");
		m.addActionListener(listener); 
		add(m); 	
		m = new MenuItem("Stop Monitor");
		m.addActionListener(listener); 	
		add(m); 
		m = new MenuItem("Show Info");
		m.addActionListener(listener); 
		add(m);
		m = new MenuItem("Update host");
		m.addActionListener(listener); 	
		add(m); 	
		
		enableEvents (AWTEvent.MOUSE_EVENT_MASK);
    } 
}


class OkCancelBox extends Dialog implements ActionListener {
 	   	 
    String msg;
    Graph graph;  	 
    
    public OkCancelBox(Frame parent, String title, String msg) { 
		super(parent, title, true);  
		graph = (Graph)parent;	  
		this.msg = msg;
		setFont(new Font("sanserif", Font.PLAIN, 12));
		
		Panel p = new Panel();  	
		Button ok = new Button("OK"); 	
		ok.addActionListener(this);
		p.add(ok);
		Button cancel = new Button("Cancel"); 	 
		cancel.addActionListener(this);
		p.add(cancel);
		add("South", p); 
		
		setSize(240, 120);
		setResizable(false); 
		show(); 	 
    }
 	
    public void paint(Graphics g) { 	
		g.drawString(msg, 10, 60);	 
    }

    public void actionPerformed(ActionEvent e) {
		String label = ((Button)e.getSource()).getLabel(); 		
		if (label.equals("OK")) { 	
			graph.graphPanel.isOK = true;
			dispose(); 
		}	
		if (label.equals("Cancel")) {
			graph.graphPanel.isOK = false;
			dispose(); 
		}
    } 	
}    
class StartProcDialog extends Dialog 
    implements ActionListener, ItemListener {

    Choice proc1; 	
     	
	 
    public StartProcDialog(Frame parent) { 
		super(parent, "Start Proc", true);  	
		Graph  graph = (Graph)parent;
		setFont(new Font("sanserif", Font.PLAIN, 12));
	
		GridBagLayout g = new GridBagLayout();
		setLayout(g);
	
		GridBagConstraints gbc = new GridBagConstraints(); 
			
		Label l; 
		l = new Label("Proc 1"); 
		gbc.weightx = 1.0; 
		gbc.weighty = 1.0; 	
		gbc.gridx = 1; 	
		gbc.gridy = 1;
		gbc.gridheight = 1; 
		gbc.gridwidth = 1;  	
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		proc1 = new Choice(); 
		proc1.addItemListener(this);
		gbc.gridx = 2; 	
		gbc.gridwidth = 2; 
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(proc1, gbc); 
		add(proc1); 

		
		Button ok = new Button("OK");  	
		ok.addActionListener(this);
		gbc.gridx = 3; 	
		gbc.gridy = 3;
		g.setConstraints(ok, gbc); 
		add(ok);  		

		Button cancel = new Button("Cancel");  	
		cancel.addActionListener(this);
		gbc.gridx = 4; 	
		g.setConstraints(cancel, gbc); 
		add(cancel);  
	
		// initialize the dialog box
		proc1.addItem("      ");
		
		for (Enumeration idEnum = graph.graphPanel.nodes.keys(); 
			 idEnum.hasMoreElements(); ) {
			String name = ((Integer)idEnum.nextElement()).toString();
		    proc1.addItem(name);
		    
		}		
		if (graph.graphPanel.pnode != null) { 			
		    proc1.select(String.valueOf(graph.graphPanel.pnode.id)); 
		   
		}

		setSize(280, 140); 	
		setResizable(false);
		show();
	}
  			
    public void actionPerformed(ActionEvent e) {
	String label = ((Button)e.getSource()).getLabel(); 		
	if (label.equals("OK")) {  	
		
		
		
		Proc p = null;
          if((p = (Proc)Engine.getLocalProc(new 				Integer(proc1.getSelectedItem()).intValue())) != null) {
            p.start();
          }
          else {
            System.out.println("Can't start PROC: PROC does not exist");
          }
	
	    dispose(); 
	}
	if (label.equals("Cancel")) 
	    dispose(); 
    }
	
    public void itemStateChanged(ItemEvent e) {}
	
} 

class PreferDialog extends Dialog implements ActionListener {
    
    TextField   box_width; 	
    TextField	box_height;
    TextField	box_dist;
    Checkbox    auto;
    Checkbox    redraw;
    Graph       graph;

    public PreferDialog(Frame parent) { 
		super(parent, "Preferences", true); 
		graph = (Graph)parent ;
		setFont(new Font("sanserif", Font.PLAIN, 12));
		setBackground(Color.lightGray);
		
		GridBagLayout g = new GridBagLayout();
		setLayout(g);	
		GridBagConstraints gbc = new GridBagConstraints(); 
		
		Label l; 
		l = new Label("Box size for each host"); 
		gbc.insets = new Insets(10, 10, 0, 0);
		gbc.weightx = 1.0; 
		gbc.weighty = 1.0; 	
		gbc.gridx = 1; 	
		gbc.gridy = 1;
		gbc.gridheight = 1; 
		gbc.gridwidth = 1;  	
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		l = new Label("Width:"); 
		gbc.insets = new Insets(0, 0, 0, 10);
		gbc.gridx = 1; 		
		gbc.gridy = 2;	
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 
		
		box_width = new TextField(3); 		
		gbc.gridx = 2; 		
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(box_width, gbc); 
		add(box_width); 
		
		l = new Label("Height:"); 
		gbc.insets = new Insets(0, 0, 0, 10);
		gbc.gridx = 3; 		
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 
		
		box_height = new TextField(3); 		
		gbc.gridx = 4; 		
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(box_height, gbc); 
		add(box_height); 
		
		l = new Label("Distance between boxes:"); 
		gbc.insets = new Insets(0, 10, 0, 0);
		gbc.gridx = 1; 		
		gbc.gridy = 3;	
		gbc.gridwidth = 3;
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(l, gbc); 
		add(l); 
		
		box_dist = new TextField(3); 		
		gbc.insets = new Insets(0, 0, 0, 0);
		gbc.gridx = 3; 		
		gbc.gridwidth = 1;
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(box_dist, gbc); 
		add(box_dist); 
		
		auto = new Checkbox("Automatically arrange nodes");
		gbc.insets = new Insets(0, 10, 0, 0);
		gbc.gridx = 1; 		
		gbc.gridy = 4;	
		gbc.gridwidth = 4;
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(auto, gbc); 
		add(auto); 
		
		redraw = new Checkbox("Redraw all", true);
		gbc.gridy = 5;	
		g.setConstraints(redraw, gbc); 
		add(redraw); 
		
		Button ok = new Button("OK");  	
		ok.addActionListener(this);
		gbc.insets = new Insets(10, 0, 10, 0);
		gbc.gridx = 2; 	
		gbc.gridy = 6;
		g.setConstraints(ok, gbc); 
		add(ok);  		
		
		Button cancel = new Button("Cancel");  	
		cancel.addActionListener(this);	 	
		gbc.gridx = 3;
		g.setConstraints(cancel, gbc); 
		add(cancel);   	
		
		// initialize
		box_width.setText(String.valueOf(graph.graphPanel.box_width)); 	
		box_height.setText(String.valueOf(graph.graphPanel.box_height)); 
		box_dist.setText(String.valueOf(graph.graphPanel.box_dist)); 
		auto.setState(graph.graphPanel.auto);
		
		//setSize(300, 200);
		pack();
		setResizable(false); 
		show();
    } 

    public void actionPerformed(ActionEvent e) {
		String label = ((Button)e.getSource()).getLabel(); 	
		if (label.equals("OK")) {
			int w = Integer.parseInt(box_width.getText());
			if (w < graph.graphPanel.MIN_BOX_WIDTH) {
				MessageBox msgBox = new MessageBox((Frame)graph, "Alert", 
							   		   "The width of box has to be greater than "
												   + graph.graphPanel.MIN_BOX_WIDTH);
				return;
			}
			else graph.graphPanel.box_width = w;
			
			int h = Integer.parseInt(box_height.getText());
			if (h < graph.graphPanel.MIN_BOX_HEIGHT) {
				MessageBox msgBox = new MessageBox((Frame)graph, "Alert", 
									 "The height of box has to be greater than "
												   +graph.graphPanel.MIN_BOX_HEIGHT);
				return;
	    }
	    else graph.graphPanel.box_height = h;

	    int dist = Integer.parseInt(box_dist.getText());
	    int min_dist = Math.max(w, h) + 40;
	    if (dist < min_dist) {
			MessageBox msgBox = new MessageBox((Frame)graph, "Alert", 
   				   "The distance between two boxes has to be greater than "+min_dist);
			return;
	    }
	    else graph.graphPanel.box_dist = dist;
	    
	    graph.graphPanel.auto = auto.getState();
		
		if (redraw.getState())
			graph.graphPanel.redrawAll();

	    dispose(); 
		}
		if (label.equals("Cancel")) 
			dispose(); 
    }

} 


class UpdateDialog extends Dialog implements ActionListener { 

    Choice host; 	
    Checkbox all;

    public UpdateDialog(Frame parent) { 
		super(parent, "Update host", true); 
		Graph graph = (Graph)parent;
		setFont(new Font("sanserif", Font.PLAIN, 12));
		setBackground(Color.lightGray);
		
		GridBagLayout g = new GridBagLayout();
		setLayout(g);	
		GridBagConstraints gbc = new GridBagConstraints(); 
		
		Label l = new Label("Hostname"); 
		gbc.insets = new Insets(10, 10, 0, 0);
		gbc.weightx = 1.0; 
		gbc.weighty = 1.0; 	
		gbc.gridx = 1; 	
		gbc.gridy = 1;
		gbc.gridheight = 1; 
		gbc.gridwidth = 1;  		
		g.setConstraints(l, gbc); 
		add(l); 
		
		host = new Choice();
		gbc.insets = new Insets(10, 0, 0, 0);
		gbc.gridx = 2; 	
		gbc.gridwidth = 2; 
		g.setConstraints(host, gbc); 
		add(host); 
		
		all = new Checkbox("Both local and remote procs", true);
		gbc.insets = new Insets(0, 10, 0, 0);
		gbc.gridx = 1; 	
		gbc.gridy = 2;	
		gbc.gridwidth = 3;  		
		g.setConstraints(all, gbc); 
		add(all); 
		
		Button ok = new Button("OK");  	
		ok.addActionListener(this);
		gbc.insets = new Insets(10, 0, 10, 0);
		gbc.gridx = 1; 		
		gbc.gridy = 3;
		//gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(ok, gbc); 
		add(ok);  		
		
		Button cancel = new Button("Cancel");  		
		cancel.addActionListener(this);
		gbc.gridx = 3; 	
		//gbc.anchor = GridBagConstraints.CENTER;
		g.setConstraints(cancel, gbc); 
		add(cancel);  	 
		
		// initialize
		host.addItem("all");
		for (Enumeration hostnameEnum = graph.graphPanel.hosts.keys(); 
			 hostnameEnum.hasMoreElements(); ) 
			host.addItem((String)hostnameEnum.nextElement());
		
		//setSize(240, 140);
		pack();
		setResizable(false); 
		show();
	} 
	
	public void actionPerformed(ActionEvent e) {
		String label = ((Button)e.getSource()).getLabel(); 	
		if (label.equals("OK")) {  	
			int i = 0;
			if (all.getState())
				i = 1;
			Engine.update(host.getSelectedItem(), i);
			dispose();  		
		}
		if (label.equals("Cancel")) 
			dispose(); 
	}
} 


class LoadClassDialog extends Dialog 
	implements ActionListener { 

    TextField  classname; 	
    Choice  host;  	 
    Graph   graph;

    public LoadClassDialog(Frame parent) { 
		super(parent, "Load class", true); 
		graph = (Graph)parent;
		setFont(new Font("sanserif", Font.PLAIN, 12));
		setBackground(Color.lightGray);
	
		GridBagLayout g = new GridBagLayout();
		setLayout(g);	
		GridBagConstraints gbc = new GridBagConstraints(); 
			
		Label l; 
		l = new Label("Load class"); 
		gbc.insets = new Insets(10, 10, 0, 0);
		gbc.weightx = 1.0; 
		gbc.weighty = 1.0; 		
		gbc.gridx = 1; 
		gbc.gridy = 1;
		gbc.gridheight = 1; 
		gbc.gridwidth = 1;  		
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		classname = new TextField(16); 		
		gbc.gridx = 2; 		
		gbc.gridwidth = 2; 
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(classname, gbc); 
		add(classname); 
	
		l = new Label("From host"); 
		gbc.insets = new Insets(10, 10, 0, 0);
		gbc.gridx = 1; 	
		gbc.gridy = 2;
		gbc.gridwidth = 1; 
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		host = new Choice();
		gbc.gridx = 2; 	
		gbc.gridy = 2;
		gbc.gridwidth = 2; 
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(host, gbc); 
		add(host); 
		 		
		Button ok = new Button("OK");  	
		ok.addActionListener(this);
		gbc.insets = new Insets(10, 0, 10, 0);
		gbc.gridx = 1; 	
		gbc.gridy = 3;
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(ok, gbc); 
		add(ok);  
		
		Button cancel = new Button("Cancel");  	
		cancel.addActionListener(this); 	
		gbc.gridx = 2;
 		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(cancel, gbc); 
		add(cancel);  	
		 
		// initialize
		classname.setText("dacia.");
		host.addItem("Any host");
		for (Enumeration hostnameEnum = graph.graphPanel.hosts.keys(); 
			 hostnameEnum.hasMoreElements();)    
		    host.addItem((String)hostnameEnum.nextElement());	
		host.remove(Engine.hostname);
		
		//setSize(220, 140);
		pack();
		setResizable(false); 
		show();
    } 
	 
    public void actionPerformed(ActionEvent e) {
		String label = ((Button)e.getSource()).getLabel(); 	
		if (label.equals("OK")) {
        if (host.getSelectedItem().equals("Any host"))
            Engine.loadClass(classname.getText(), null);
        else
            Engine.loadClass(classname.getText(), host.getSelectedItem()); 	   
        dispose(); 
		}
		if (label.equals("Cancel")) 
			dispose(); 
    }
	
} 


class ConnectProcsDialog extends Dialog 
    implements ActionListener, ItemListener {

    Choice proc1; 	
    Choice port1; 
    Choice proc2; 
    Choice port2; 	
	 
    public ConnectProcsDialog(Frame parent) { 
		super(parent, "Connect Procs", true);  	
		Graph  graph = (Graph)parent;
		setFont(new Font("sanserif", Font.PLAIN, 12));
		setBackground(Color.lightGray);

		GridBagLayout g = new GridBagLayout();
		setLayout(g);
	
		GridBagConstraints gbc = new GridBagConstraints(); 
			
		Label l; 
		l = new Label("Proc 1"); 
		gbc.insets = new Insets(10, 10, 0, 0);
		gbc.weightx = 1.0; 
		gbc.weighty = 1.0; 	
		gbc.gridx = 1; 	
		gbc.gridy = 1;
		gbc.gridheight = 1; 
		gbc.gridwidth = 1;  	
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		proc1 = new Choice(); 
		proc1.addItemListener(this);
		gbc.gridx = 2; 	
		gbc.gridwidth = 2; 
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(proc1, gbc); 
		add(proc1); 

		l = new Label("Port"); 
		gbc.gridx = 4; 	
		gbc.gridwidth = 1; 
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 

		port1 = new Choice();
		gbc.insets = new Insets(10, 0, 0, 10);
		gbc.gridx = 5;
 		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(port1, gbc); 
		add(port1); 
	
		l = new Label("Proc 2"); 
		gbc.insets = new Insets(10, 10, 0, 0);
		gbc.gridx = 1; 	
		gbc.gridy = 2;
		gbc.gridwidth = 1; 
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		proc2 = new Choice();
		proc2.addItemListener(this);
		gbc.gridx = 2; 	
		gbc.gridwidth = 2; 
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(proc2, gbc); 
		add(proc2); 
	
		l = new Label("Port"); 
		gbc.gridx = 4; 	
		gbc.gridwidth = 1; 
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		port2 = new Choice(); 
		gbc.insets = new Insets(10, 0, 0, 10);
		gbc.gridx = 5; 	
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(port2, gbc); 
		add(port2); 
	 	
		Button ok = new Button("OK");  	
		ok.addActionListener(this);
		gbc.insets = new Insets(10, 0, 10, 0);
		gbc.gridx = 3; 	
		gbc.gridy = 3;
		g.setConstraints(ok, gbc); 
		add(ok);  		

		Button cancel = new Button("Cancel");  	
		cancel.addActionListener(this);
		gbc.gridx = 4; 	
		g.setConstraints(cancel, gbc); 
		add(cancel);  
	
		// initialize the dialog box
		proc1.addItem("      ");
		proc2.addItem("      ");
		port1.addItem("  ");
		port2.addItem("  ");
		for (Enumeration idEnum = graph.graphPanel.nodes.keys(); 
			 idEnum.hasMoreElements(); ) {
			String name = ((Integer)idEnum.nextElement()).toString();
		    proc1.addItem(name);
		    proc2.addItem(name);
		}		
		if (graph.graphPanel.pnode != null) { 			
		    proc1.select(String.valueOf(graph.graphPanel.pnode.id)); 
		    Util.getUnusedPorts(proc1, port1);
		}

		//setSize(240, 140); 
		pack();
		setResizable(false);
		show();
	}
  			
    public void actionPerformed(ActionEvent e) {
	String label = ((Button)e.getSource()).getLabel(); 		
	if (label.equals("OK")) {  	
		if (!Engine.connectProcs(Integer.parseInt(proc1.getSelectedItem()), 
								Integer.parseInt(port1.getSelectedItem()), 
								Integer.parseInt(proc2.getSelectedItem()), 
								Integer.parseInt(port2.getSelectedItem()), true) ) {
			MessageBox msgBox = new MessageBox((Frame)getParent(), 
											   "Alert", "Invalid input!");  
		}
	    dispose(); 
	}
	if (label.equals("Cancel")) 
	    dispose(); 
    }
	
    public void itemStateChanged(ItemEvent e) {
		Object src = e.getSource();
		if (src == proc1) 
			Util.getUnusedPorts(proc1, port1);
		else if (src == proc2)
			Util.getUnusedPorts(proc2, port2); 
    }
	
} 

class DisconnectProcDialog extends Dialog 
    implements ActionListener, ItemListener { 

    Choice  proc; 	
    Choice  port;  

    public DisconnectProcDialog(Frame parent) { 
		super(parent, "Disconnect Proc", true); 
		Graph graph = (Graph)parent;
		setFont(new Font("sanserif", Font.PLAIN, 12));
		setBackground(Color.lightGray);
	
		GridBagLayout g = new GridBagLayout();
		setLayout(g);	
		GridBagConstraints gbc = new GridBagConstraints(); 
			
		Label l; 
		l = new Label("Proc"); 
		gbc.insets = new Insets(10, 10, 0, 0);
		gbc.weightx = 1.0; 
		gbc.weighty = 1.0; 	
		gbc.gridx = 1; 	
		gbc.gridy = 1;
		gbc.gridheight = 1; 
		gbc.gridwidth = 1;  		
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		proc = new Choice();
 		proc.addItemListener(this);
		gbc.gridx = 2; 	
		gbc.gridwidth = 2; 
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(proc, gbc); 
		add(proc); 
	
		l = new Label("Port"); 
		gbc.gridx = 4; 	
		gbc.gridy = 1;
		gbc.gridwidth = 1; 
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		port = new Choice();
		gbc.insets = new Insets(10, 0, 0, 10);
		gbc.gridx = 5;
		gbc.weightx = 0.6;  		
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(port, gbc); 
		add(port); 
		 		
		Button ok = new Button("OK");  	
		ok.addActionListener(this);
		gbc.insets = new Insets(10, 0, 10, 0);
		gbc.gridx = 3; 		
		gbc.gridy = 2;
		gbc.anchor = GridBagConstraints.CENTER;
		g.setConstraints(ok, gbc); 
		add(ok);  		
		
		Button cancel = new Button("Cancel");  		
		cancel.addActionListener(this); 		
		gbc.gridx = 4; 	
		gbc.anchor = GridBagConstraints.CENTER;
		g.setConstraints(cancel, gbc); 
		add(cancel);  

		// initialize
		proc.addItem("      ");
		port.addItem("  ");
		for (Enumeration idEnum = graph.graphPanel.nodes.keys(); 
			 idEnum.hasMoreElements();) 
		    proc.addItem(((Integer)idEnum.nextElement()).toString());		
		if (graph.graphPanel.pnode != null) {		
		    proc.select(String.valueOf(graph.graphPanel.pnode.id)); 
		    Util.getUsedPorts(proc, port);
		}
		else proc.select("      ");
		 	
		//setSize(240, 120);
		pack();
		setResizable(false); 
		show();
	} 		 	

    public void actionPerformed(ActionEvent e) {			
	    String label = ((Button)e.getSource()).getLabel(); 	
      if (label.equals("OK")) {  	
          if (!Engine.disconnectProcs(Integer.parseInt(proc.getSelectedItem()), 
                      Integer.parseInt(port.getSelectedItem()), true)) {
              MessageBox msgBox = new MessageBox((Frame)getParent(), 
                                                 "Alert", "Invalid input!"); 		
          }	   
          dispose(); 
      }
      if (label.equals("Cancel")) 
          dispose(); 		 
    }
	
    public void itemStateChanged(ItemEvent e) {
        Object src = e.getSource();
        if (src == proc) 
            Util.getUsedPorts(proc, port);
    }
		 
} 

class MoveProcDialog extends Dialog 
	implements ActionListener, ItemListener { 

    Choice  proc; 	
    Choice  host;  	 
    Graph   graph;

    public MoveProcDialog(Frame parent) { 
		super(parent, "Move Proc", true); 
		graph = (Graph)parent;
		setFont(new Font("sanserif", Font.PLAIN, 12));
		setBackground(Color.lightGray);
	
		GridBagLayout g = new GridBagLayout();
		setLayout(g);	
		GridBagConstraints gbc = new GridBagConstraints(); 
			
		Label l; 
		l = new Label("ProcID"); 
		gbc.insets = new Insets(10, 10, 0, 0);
		gbc.weightx = 0.8; 
		gbc.weighty = 1.0; 		
		gbc.gridx = 1; 
		gbc.gridy = 1;
		gbc.gridheight = 1; 
		gbc.gridwidth = 1;  		
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		proc = new Choice(); 
		proc.addItemListener(this);
		gbc.insets = new Insets(10, 0, 10, 0);
		gbc.gridx = 2; 	
		gbc.gridwidth = 2; 
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(proc, gbc); 
		add(proc); 
	
		l = new Label("Hostname"); 
		gbc.insets = new Insets(10, 10, 0, 0);
		gbc.gridx = 1; 	
		gbc.gridy = 2;
		gbc.gridwidth = 1; 
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		host = new Choice();
		gbc.insets = new Insets(10, 0, 0, 10);
		gbc.gridx = 2; 	
		gbc.gridy = 2;
		gbc.gridwidth = 2; 
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(host, gbc); 
		add(host); 
		 		
		Button ok = new Button("OK");  	
		ok.addActionListener(this);
		gbc.insets = new Insets(10, 0, 10, 0);
		gbc.gridx = 1; 	
		gbc.gridy = 3;
		gbc.anchor = GridBagConstraints.CENTER;
		g.setConstraints(ok, gbc); 
		add(ok);  
		
		Button cancel = new Button("Cancel");  	
		cancel.addActionListener(this); 	
		gbc.gridx = 3;
 		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(cancel, gbc); 
		add(cancel);  	
		 
		// initialize
		proc.addItem("      ");
		host.addItem("        ");
		for (Enumeration idEnum = graph.graphPanel.nodes.keys(); 
			 idEnum.hasMoreElements();) 
			proc.addItem(((Integer)idEnum.nextElement()).toString()); 	
		for (Enumeration hostnameEnum = graph.graphPanel.hosts.keys(); 
			 hostnameEnum.hasMoreElements();) 
		    host.addItem((String)hostnameEnum.nextElement());	
		if (graph.graphPanel.pnode != null) {		
		    proc.select(String.valueOf(graph.graphPanel.pnode.id)); 
			host.remove(graph.graphPanel.pnode.host);
		}

		//setSize(220, 140);
		pack();
		setResizable(false); 
		show();
	} 
	 
    public void actionPerformed(ActionEvent e) {
		String label = ((Button)e.getSource()).getLabel(); 	
		if (label.equals("OK")) {  	
			if (!Engine.moveProc(Integer.valueOf(proc.getSelectedItem()),  	
								 host.getSelectedItem())) {	
				MessageBox msgBox = new MessageBox((Frame)getParent(),
												   "Alert", "Invalid input!"); 	
			}				   
			dispose(); 
		}
		if (label.equals("Cancel")) 
			dispose(); 
    }
	
    public void itemStateChanged(ItemEvent e) {
		Object src = e.getSource();
		if (src == proc) {
			Integer procID = Integer.valueOf(proc.getSelectedItem());
			Node n = (Node)graph.graphPanel.nodes.get(procID);
			if (n == null)
				return;
			host.removeAll();
			for (Enumeration hostnameEnum = graph.graphPanel.hosts.keys(); 
				 hostnameEnum.hasMoreElements(); ) {
				String name = (String)hostnameEnum.nextElement();
				if (!name.equals(n.host))
					host.addItem(name);	
			}
		}
	}


} 


class RemoveProcDialog extends Dialog 
	implements ActionListener { 

    Choice  proc; 	
    Graph   graph;

    public RemoveProcDialog(Frame parent) { 
		super(parent, "Remove Proc", true); 
		graph = (Graph)parent;
		setFont(new Font("sanserif", Font.PLAIN, 12));
		setBackground(Color.lightGray);
	
		GridBagLayout g = new GridBagLayout();
		setLayout(g);	
		GridBagConstraints gbc = new GridBagConstraints(); 
			
		Label l; 
		l = new Label("ProcID"); 
		gbc.insets = new Insets(10, 10, 0, 0);
		gbc.weightx = 0.8; 
		gbc.weighty = 1.0; 		
		gbc.gridx = 1; 
		gbc.gridy = 1;
		gbc.gridheight = 1; 
		gbc.gridwidth = 1;  		
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		proc = new Choice(); 
		gbc.insets = new Insets(10, 0, 0, 10);
		gbc.gridx = 2; 	
		gbc.gridwidth = 2; 
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(proc, gbc); 
		add(proc); 
		 		
		Button ok = new Button("OK");  	
		ok.addActionListener(this);
		gbc.insets = new Insets(10, 0, 10, 0);
		gbc.gridx = 1; 	
		gbc.gridy = 3;
		gbc.anchor = GridBagConstraints.CENTER;
		g.setConstraints(ok, gbc); 
		add(ok);  
		
		Button cancel = new Button("Cancel");  	
		cancel.addActionListener(this); 	
		gbc.gridx = 3;
 		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(cancel, gbc); 
		add(cancel);  	
		 
		// initialize
		proc.addItem("      ");
		for (Enumeration idEnum = graph.graphPanel.nodes.keys(); 
			 idEnum.hasMoreElements();) 
			proc.addItem(((Integer)idEnum.nextElement()).toString()); 	
		if (graph.graphPanel.pnode != null) {		
		    proc.select(String.valueOf(graph.graphPanel.pnode.id)); 
		}

		//setSize(220, 100);
		pack();
		setResizable(false); 
		show();
    } 
	 
    public void actionPerformed(ActionEvent e) {
		String label = ((Button)e.getSource()).getLabel(); 	
		if (label.equals("OK")) {  	
			if (!Engine.removeProc(Integer.valueOf(proc.getSelectedItem()))) {	
				MessageBox msgBox = new MessageBox((Frame)getParent(),
												   "Alert", "Invalid input!"); 	
			}				   
			dispose(); 
		}
		if (label.equals("Cancel")) 
			dispose(); 
    }
	
} 

class ConnectEngineDialog extends Dialog implements ActionListener { 

    TextField   host; 	
    TextField	port;  	

    public ConnectEngineDialog(Frame parent) { 
		super(parent, "Connect to Engine", true); 
		Graph graph = (Graph)parent ;
		setFont(new Font("sanserif", Font.PLAIN, 12));
		setBackground(Color.lightGray);
	
		GridBagLayout g = new GridBagLayout();
		setLayout(g);
	
		GridBagConstraints gbc = new GridBagConstraints(); 
			
		Label l; 
		l = new Label("Hostname"); 
		gbc.insets = new Insets(10, 10, 0, 0);
		gbc.weightx = 1.0; 
		gbc.weighty = 1.0; 	
		gbc.gridx = 1; 	
		gbc.gridy = 1;
		gbc.gridheight = 1; 
		gbc.gridwidth = 1;  	
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		host = new TextField(10); 		
		gbc.insets = new Insets(10, 0, 0, 10);
		gbc.gridx = 2; 		
		gbc.gridwidth = 2; 
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(host, gbc); 
		add(host); 

		l = new Label("Port"); 
		gbc.insets = new Insets(10, 10, 0, 0);
		gbc.gridx = 1; 	
		gbc.gridy = 2;
		gbc.gridwidth = 1; 
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(l, gbc); 
		add(l); 

		port = new TextField(4); 	
		gbc.insets = new Insets(10, 0, 0, 10);
		gbc.gridx = 2; 		
		gbc.gridwidth = 2; 
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(port, gbc); 
		add(port); 

 		Button ok = new Button("OK");  	
		ok.addActionListener(this);
		gbc.insets = new Insets(10, 0, 10, 0);
		gbc.gridx = 1; 	
		gbc.gridy = 3;
		gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(ok, gbc); 
		add(ok);  		

		Button cancel = new Button("Cancel");  	
		cancel.addActionListener(this);	 	
		gbc.gridx = 3;
 		gbc.anchor = GridBagConstraints.CENTER;
		g.setConstraints(cancel, gbc); 
		add(cancel);   	
	
		// initialize
		port.setText((new Integer(Engine.port)).toString()); 	
		host.requestFocus();

		//setSize(200, 140);
		pack();
		setResizable(false); 
		show();
	} 

    public void actionPerformed(ActionEvent e) {
		String label = ((Button)e.getSource()).getLabel(); 	
		if (label.equals("OK")) {  			
			if (Engine.connect(host.getText(), 
							   Integer.parseInt(port.getText()), true) == null) { 	
				MessageBox msgBox = new MessageBox((Frame)getParent(), 
												   "Alert", "Invalid input!"); 	
			}				   
			dispose(); 
		}
		if (label.equals("Cancel")) 
			dispose(); 
    }
} 

/**
 * The graph panel is the area in which the graph of the application components 
 * are displayed
 */
class GraphPanel extends Panel
    implements Runnable, MouseListener, MouseMotionListener {

    private Dimension sizeMin;   // dimension of the graph
    Dimension sizeVP;    // size of the viewport
 
    int nhosts;     
    Hashtable hosts = new Hashtable(10);

    int nnodes;
    Hashtable nodes = new Hashtable(100);

    int ncables;
    Vector cables = new Vector(10, 10);

    int nedges;
    Vector edges = new Vector(10, 10);

    int ncolors;
 
    private Thread relaxer;


    static final int TOLERANCE = 6;
    static final int MIN_BOX_WIDTH  = 80;
    static final int MIN_BOX_HEIGHT = 80;
    static final int MIN_GRAPH_WIDTH  = 400;
    static final int MIN_GRAPH_HEIGHT = 300;
    static final int MAX_COLOR = 12;

    static final Color colors[] = {Color.white,
		      Color.getHSBColor((float)0.15, (float)0.3, (float)1.0), 
		      Color.getHSBColor((float)0.5, (float)0.3, (float)1.0), 
		      Color.getHSBColor((float)0.0, (float)0.2, (float)1.0), 
		      Color.getHSBColor((float)0.25, (float)0.3, (float)1.0), 
		      Color.getHSBColor((float)0.05, (float)0.5, (float)1.0), 
		      Color.getHSBColor((float)0.45, (float)0.7, (float)0.9),
		      Color.getHSBColor((float)0.6, (float)0.2, (float)1.0), 
		      Color.getHSBColor((float)0.6, (float)0.6, (float)1.0), 
		      Color.getHSBColor((float)0.75, (float)0.4, (float)1.0), 
		      Color.getHSBColor((float)0.85, (float)0.3, (float)1.0), 
		      Color.getHSBColor((float)0.95, (float)0.6, (float)1.0)}; 

    int box_width  = 200;
    int box_height = 200;
    int box_dist   = 250;
    boolean  auto;
    boolean  isOK;

    GraphPanel() {
		  setFont(new Font("sanserif", Font.PLAIN, 12));
		  ncolors = 0;
		  auto = false;
		  sizeMin = new Dimension(400, 300);
		  addMouseListener(this); 
		  addMouseMotionListener(this);
    }

	  // for debug purposes
    private void dump() { 
		for (Enumeration hostEnum=hosts.elements(); hostEnum.hasMoreElements(); ) {
			Host h = (Host)hostEnum.nextElement();		
			System.out.println("host " + h.name + " " + h.color);
		} 
		for (Enumeration nodeEnum=nodes.elements(); nodeEnum.hasMoreElements(); ) {
 			Node n = (Node)nodeEnum.nextElement();		
			System.out.println("node " + n.id);
		}
		System.out.println("ncables = "+ncables);
		System.out.println("nedges  = "+nedges+ "\n");
    }
	
		
	/**
	 * Updates the graph when the configuration changes.
	 */
    public void engineUpdate() {
        //long t1 = System.currentTimeMillis();

		Hashtable	proclist;
		Enumeration procEnum;  
		Proc		proc;
		RemoteProc	rmproc;
		
		nedges = 0;
		ncables = 0;
	 	
		edges.removeAllElements();
		cables.removeAllElements();
		
		// add local Procs 	
		proclist = Engine.getLocalProcs();
		procEnum = proclist.elements();
		while (procEnum.hasMoreElements()) {
			proc = (Proc)procEnum.nextElement();
			addNode(proc.id, proc.hostname, proc.name, true);
			for (int i=0; i<proc.portNo; i++) {
				if (proc.port[i].connectionStatus != 0) 
					addEdge(proc.id, proc.port[i].connectedProcID, 80); 
			}
		}
		
		// add remote Procs 	
		proclist = Engine.getRemoteProcs();
		procEnum = proclist.elements();
		while (procEnum.hasMoreElements()) {
			rmproc = (RemoteProc)procEnum.nextElement();
			addNode(rmproc.id, rmproc.hostname, rmproc.name, false);
			for (int i=0; i<rmproc.portNo; i++) {
				if (rmproc.connectedProcID[i] != 0) 
					addEdge(rmproc.id, rmproc.connectedProcID[i], 80); 
			}
		} 	 	

		// add cables
		for(int i = 0; i < Engine.connections.size(); i++) {
			Socket s = ((Connection)Engine.connections.elementAt(i)).socket;
			String hostname = s.getInetAddress().getHostName().toLowerCase(); 	
			int j = hostname.indexOf(".");
			if(j > 0)
				hostname = hostname.substring(0, j); 	
			addCable(Engine.hostname, hostname);
		}		

    //long dt = System.currentTimeMillis() - t1;
    //System.out.println("engineUpdate() " + dt);
    }

	/**
	 * Regenerates the graph. This function is more time-consuming than 
	 * engineUpdate().
	 */
    public void redrawAll() {
		long t1 = System.currentTimeMillis();

		Hashtable	proclist;
		Enumeration procEnum;  
		Proc		proc;
		RemoteProc	rmproc;
		
		nhosts = 0;
		nnodes = 0;
		nedges = 0;	
		ncables = 0;
		ncolors = 0;
	
		hosts.clear();
		nodes.clear();
		edges.removeAllElements();
		cables.removeAllElements();
		
		// add local nodes
		proclist = Engine.getLocalProcs();
    	procEnum = proclist.elements();
    	while (procEnum.hasMoreElements()) {
			proc = (Proc)procEnum.nextElement();
			addNode(proc.id, proc.hostname, proc.name, true);
			for (int i=0; i<proc.portNo; i++) {
				if (proc.port[i].connectionStatus != 0) 
					addEdge(proc.id, proc.port[i].connectedProcID, 80); 
			}
		}
		
		// add remote nodes
		proclist = Engine.getRemoteProcs();
    	procEnum = proclist.elements();
    	while (procEnum.hasMoreElements()) {
			rmproc = (RemoteProc)procEnum.nextElement();
			addNode(rmproc.id, rmproc.hostname, rmproc.name, false);
			for (int i=0; i<rmproc.portNo; i++) {
				if (rmproc.connectedProcID[i] != 0) 
					addEdge(rmproc.id, rmproc.connectedProcID[i], 80); 
			}
		} 	 	
		
		// add cables
		for(int i = 0; i < Engine.connections.size(); i++) {
			Socket s = ((Connection)Engine.connections.elementAt(i)).socket;
			String hostname = s.getInetAddress().getHostName().toLowerCase(); 	
			int j = hostname.indexOf(".");
			if(j > 0)
				hostname = hostname.substring(0, j); 	
			addCable(Engine.hostname, hostname);
		}
		
		// arrange all the hosts
		Host h = (Host)hosts.get(Engine.hostname);
		if (hosts.size() <= 4)
			h.x = 50;
		else h.x = 150;
		int x = h.x;
		int y = h.y = 50;
		h.w = box_width;
		h.h = box_height;
		
		double angle = 2*Math.PI/nhosts; 
		
		int k = 0;
		for (Enumeration hostEnum = hosts.elements(); hostEnum.hasMoreElements(); ) {
			h = (Host)hostEnum.nextElement();
			if (!h.name.equals(Engine.hostname)) {
				h.x = x + (int)(box_dist*Math.cos(k*angle)); 	
				h.y = y + (int)(box_dist*Math.sin(k*angle));
				h.w = box_width;
				h.h = box_height;
				x = h.x;
				y = h.y;
				k++;
			}
		} 	

		// rearrange all the nodes
		for (Enumeration nodeEnum = nodes.elements(); nodeEnum.hasMoreElements(); ) {
			Node n = (Node)nodeEnum.nextElement();
			h = (Host)hosts.get(n.host);
			if (!isNodeInsideHost(n, h)) {
				n.x = h.x + n.w/2 + (int)((h.w-n.w) * Math.random());
				n.y = h.y + n.h/2 + (int)((h.h-n.h) * Math.random());
			}
		}		
		
		getMinSize();
		
		//long dt = System.currentTimeMillis() - t1;
		//System.out.println("redrawAll() " + dt);
    }


    /**
	 * Adds a host
	 */
    private Host addHost(String name) { 
		Host h = (Host)hosts.get(name); 	
		if (h != null) 		
			return h; 		 
		
		h = new Host();
		h.name = name;
		h.w = box_width;
		h.h = box_height;
		h.color = ncolors++;
		if (ncolors >= MAX_COLOR) 
			ncolors = 1;
		
		// determine the positon of the new host
		if (nhosts == 0) { // local host
			h.x = 50;
			h.y = 50;
		}
		else { // remote host
			Host h1;
			String bottom = null;   // the index of the host with largest Y
			int maxY = 0;           // maxY is the largest Y
			int maxY1 = 0;          // maxY1 is the second largest Y
			
			for (Enumeration hostEnum = hosts.elements(); 
				 hostEnum.hasMoreElements(); ) {
				h1 = (Host)hostEnum.nextElement();			
				if (maxY <= h1.y + h1.h) {
					maxY1 = maxY;
					maxY = h1.y + h1.h;
					bottom = h1.name;
				}
			}
			h.y = maxY1 + 50;
			h1 = (Host)hosts.get(bottom);
			if (h1.x > box_width+80) 
				// has room at left of the bottom host
				h.x = h1.x - box_width - 50;
			else if (h1.x+h1.w+box_width+50 < sizeVP.width)
				// has room at the right of the bottom host
				h.x = h1.x + h1.w + 50;
			else {			
				h.y = maxY + 30;
				h.x = (int) ((sizeVP.width - box_width -50) * Math.random()) + 50;
			} 
		}
		
		hosts.put(name, h);
		nhosts++;
		getMinSize(); 
		return h;
    }	 		

    /**
	 * Removes a host
	 * @param name  the name of the host to be removed
	 */
    void removeHost(String name) {
		// remove the host
		hosts.remove(name);
		nhosts--;
		
		// remove the cables connected to the host
		for (int i = 0; i < cables.size(); i++) {
			Cable c = (Cable)cables.elementAt(i);
			if (c.from.equals(name) || c.to.equals(name)) {
				cables.removeElementAt(i);
				i--;
				ncables--;
			}
		}
		
		// remove the nodes in the host
		for (Enumeration nodeEnum = nodes.elements(); nodeEnum.hasMoreElements(); ) {
			Node n = (Node)nodeEnum.nextElement();	
			if (n.host.equals(name)) {
				removeNode(n.id);
			}
		}
	}  

    /**
	 * Removes a PROC
	 * @param name  the ID of the PROC to be removed
	 */
    void removeNode(int id) {
		// remove the node
		nodes.remove(new Integer(id));
		nnodes--;

		// remove the edges connected to the node
		for (int i = 0; i < edges.size(); i++) {
			Edge e = (Edge)edges.elementAt(i);
			if (e.from == id || e.to == id) {
				edges.removeElementAt(i);
				i--;
				nedges--;
			}
		}
	}
      
    /**
	 * Adds a connection between hosts
	 */
    private void addCable(String from, String to) {
		Cable c = new Cable();
		addHost(from);
		addHost(to);
		c.from = from;
		c.to  = to;
		cables.addElement(c);
		ncables++;
    }

    /**
	 * Adds a PROC
	 */
    private Node addNode(int id) {
		Hashtable	proclist;
		Proc		proc;
		RemoteProc	rmproc;
		
		proclist = Engine.getLocalProcs();
		proc = (Proc)proclist.get(new Integer(id));
		if (proc != null) 
			return addNode(proc.id, proc.hostname, proc.name, true); 
		
		proclist = Engine.getRemoteProcs();
		rmproc = (RemoteProc)proclist.get(new Integer(id));
		if (rmproc != null) 
			return addNode(rmproc.id, rmproc.hostname, proc.name, false); 
		
		return null;
    }

    /**
	 * Adds a PROC
	 */
    private Node addNode(int id, String hostname, String procname, boolean local) {  
		Node n = (Node)nodes.get(new Integer(id));
		if (n != null) {
			n.local = local; 
			n.host = hostname;
			n.name = procname; 
			Host h = (Host)hosts.get(n.host);
			if (h != null && !isPointInsideHost(n.x, n.y, h)) { 
				// the proc is moved to another host
				n.x = h.x + n.w/2 + (int)((h.w-n.w) * Math.random());
				n.y = h.y + n.h/2 + (int)((h.h-n.h) * Math.random());
			}
			return n;
		}
		
		//	System.out.println("add proc " + proc.id);  
		n = new Node();
		Host h = addHost(hostname); 
		n.host = hostname;
		n.local = local; 
		n.id = id; 
		n.name = procname; 
		
		FontMetrics fm = getFontMetrics(getFont()); 	    
		n.w = Math.max(fm.stringWidth(String.valueOf(id)), 
					   fm.stringWidth(procname)) + 6;
		n.h = 2*fm.getHeight();
		n.x = h.x + n.w/2 + (int)((h.w-n.w) * Math.random());
		n.y = h.y + n.h/2 + (int)((h.h-n.h) * Math.random());
		
		nodes.put(new Integer(n.id), n);
		nnodes++;
		return n;
    }

    /**
	 * Adds a connection between nodes
	 */
    private void addEdge(int from, int to, int len) {
		Edge e;
		Node n1 = (Node)nodes.get(new Integer(from));
		Node n2 = (Node)nodes.get(new Integer(to));
		if (n1 == null || n2 == null)
			return;
		
		// look for existing edges
		for (int i = 0 ; i < nedges ; i++) {
			e = (Edge)edges.elementAt(i);
			if ((from==e.from && to==e.to) || (from==e.to && to==e.from))
				return;
		} 
		e = new Edge();
		e.from = from;
		e.to  = to;
		e.len = len;
		edges.addElement(e);
		nedges++;
    }

    public void run() {
	Thread me = Thread.currentThread();
		while (relaxer == me) {
			relax();
			try {
				Thread.sleep(100);
			} 
			catch (InterruptedException e) {
				break;
			} 
		}
    }

    private synchronized void relax() {
		if (!auto) {
			repaint();
			return;
		}
		
		for (int i = 0; i < edges.size(); i++ ) {
			Edge e  = (Edge)edges.elementAt(i);
			Node n1 = (Node)nodes.get(new Integer(e.from));
			Node n2 = (Node)nodes.get(new Integer(e.to));
			double vx = n2.x - n1.x;
			double vy = n2.y - n1.y;
			double len = Math.sqrt(vx * vx + vy * vy);
			len = (len == 0) ? .0001 : len;
			double f = (e.len - len) / (len * 3);
			double dx = f * vx;
			double dy = f * vy;
			
			n2.dx += dx;
			n2.dy += dy;
			n1.dx += -dx;
			n1.dy += -dy;
		}
		
		for (Enumeration nodeEnum1 = nodes.elements(); 
			 nodeEnum1.hasMoreElements(); ) {
			Node n1 = (Node)nodeEnum1.nextElement();	
			double dx = 0;
			double dy = 0;
			
			for (Enumeration nodeEnum2 = nodes.elements(); 
				 nodeEnum2.hasMoreElements(); ) {
				Node n2 = (Node)nodeEnum2.nextElement();	
				if (n1.id == n2.id) 
					continue;
				
				double vx = n1.x - n2.x;
				double vy = n1.y - n2.y;
				double len = vx * vx + vy * vy;
				if (len == 0) {
					dx += Math.random();
					dy += Math.random();
				} 
				else if (len < 100*100) {
					dx += vx / len;
					dy += vy / len;
				}
			}
			double dlen = dx * dx + dy * dy;
			if (dlen > 0) {
				dlen = Math.sqrt(dlen) / 2;
				n1.dx += dx / dlen;
				n1.dy += dy / dlen;
			}
		}
		
		for (Enumeration nodeEnum = nodes.elements(); nodeEnum.hasMoreElements(); ) {
			Node n = (Node)nodeEnum.nextElement();	
			Host h = (Host)hosts.get(n.host);
			n.x += Math.max(-5, Math.min(5, n.dx));
			n.y += Math.max(-5, Math.min(5, n.dy));
			if (n.x < 30 + (int)h.x) {
				n.x = 30 + (int)h.x;
			} 
			else if (n.x > 110 + (int)h.x) {
				n.x =  110 + (int)h.x;
			}
			if (n.y < 20 + (int)h.y) {
				n.y = 20 + (int)h.y;
			} 
			else if (n.y > 120 + (int)h.y) {
				n.y =  120 + (int)h.y;
			}
			n.dx /= 2;
			n.dy /= 2;
		}
		repaint();
    }
	
    /** Returns true if point (x, y) is inside Host h. */	
    private boolean isPointInsideHost(int x, int y, Host h) {
		if (x >= h.x && x <= h.x+h.w && y >= h.y && y <= h.y+h.h) 
			return true;
		return false;
	} 	 	

    /** Returns true if point (x, y) is inside Node n. */ 	
    private boolean isPointInsideNode(int x, int y, Node n) {
		if (x >= n.x-n.w/2 && x <= n.x+n.w/2 && y >= n.y-n.h/2 && y <= n.y+n.h/2) 
			return true;
		return false;
    } 
	
    /** Returns true if Node n is inside Host h. */	
    private boolean isNodeInsideHost(Node n, Host h) {
		if (isPointInsideHost(n.x-n.w/2, n.y-n.h/2, h) &&
			isPointInsideHost(n.x+n.w/2, n.y+n.h/2, h))
			return true;
		return false;
    } 

    /** Computes the minimal size of the graph. */
    private void getMinSize() {
		sizeMin = new Dimension(MIN_GRAPH_WIDTH, MIN_GRAPH_HEIGHT); 		
		for (Enumeration hostEnum = hosts.elements(); hostEnum.hasMoreElements(); ) {
			Host h = (Host)hostEnum.nextElement();
			sizeMin.width  = Math.max(h.x+h.w+30, sizeMin.width);
			sizeMin.height = Math.max(h.y+h.h+30, sizeMin.height);
		}
    }

    int px;  // picked point
    int py;

    Node pnode;			// picked node
    Host phost, resized;	// picked host
    Image offscreen;
    Graphics offgraphics;

    /** Paints the box representing host. */
    public void paintHost(Graphics g, Host h, FontMetrics fm, Color color) { 		
		g.setColor(color);
		g.fillRect(h.x, h.y, h.w, h.h);
		g.setColor(Color.black);
		g.drawRect(h.x, h.y, h.w, h.h);
		g.drawString(h.name, h.x+5, h.y-5);
    }

    /** Paints the box representing PROC. */
    public void paintNode(Graphics g, Node n, FontMetrics fm) { 
		Host host = (Host)hosts.get(n.host);
		if (host == null)
			return; 	

		int x = (int)n.x;
		int y = (int)n.y;
		int w = n.w;
		int h = n.h; 	
		g.setColor(colors[host.color]);
		g.fillRect(x - w/2, y - h / 2, w, h);
		g.setColor(Color.black);
		g.drawRect(x - w/2, y - h / 2, w, h);
		g.drawString(String.valueOf(n.id), x - (w-6)/2, y);
		g.drawString(n.name, x - (w-6)/2, y + fm.getAscent());
    }
 	
    public synchronized void update(Graphics g) {
		Dimension  sizePanel = getSize();
		if (offscreen == null 
			|| sizePanel.width < sizeMin.width 
			|| sizePanel.height < sizeMin.height 
			|| sizePanel.width > sizeMin.width && sizePanel.width > sizeVP.width
			|| sizePanel.height > sizeMin.height && sizePanel.height > sizeVP.height) {
			sizeVP = ((ScrollPane)getParent()).getViewportSize();
			sizePanel.width = Math.max(sizeVP.width, sizeMin.width);
			sizePanel.height = Math.max(sizeVP.height, sizeMin.height);
			offscreen = createImage(sizePanel.width, sizePanel.height);
			offgraphics = offscreen.getGraphics();
			offgraphics.setFont(getFont());
		}	 	 	
		offgraphics.setColor(getBackground());
		offgraphics.fillRect(0, 0, sizePanel.width, sizePanel.height); 	
		
		// draw cables 	
		for (int i = 0 ; i < ncables ; i++) {
			Cable c = (Cable)cables.elementAt(i);
			Host h1 = (Host)hosts.get(c.from);
			Host h2 = (Host)hosts.get(c.to);
			offgraphics.setColor(Color.blue);
			offgraphics.drawLine(h1.x+h1.w/2, h1.y+h1.h/2, h2.x+h2.w/2, h2.y+h2.h/2);
			offgraphics.drawLine(h1.x+h1.w/2+1, h1.y+h1.h/2, h2.x+h2.w/2+1, h2.y+h2.h/2);
			offgraphics.drawLine(h1.x+h1.w/2, h1.y+h1.h/2-1, h2.x+h2.w/2, h2.y+h2.h/2-1);
		}	 	
		
		// draw boxes
		FontMetrics fm = offgraphics.getFontMetrics();
		Enumeration hostEnum = hosts.elements();
		while (hostEnum.hasMoreElements()) {
			Host h = (Host)hostEnum.nextElement();
			if (h.name.equals(Engine.hostname))
				paintHost(offgraphics, h, fm, Color.yellow); 
			else
				paintHost(offgraphics, h, fm, Color.white); 
		}		 		
		
		// draw edges
		for (int i = 0 ; i < nedges ; i++) {
			Edge e  = (Edge)edges.elementAt(i);
			Node n1 = (Node)nodes.get(new Integer(e.from));
			Node n2 = (Node)nodes.get(new Integer(e.to));
			int x1 = (int)n1.x;
			int y1 = (int)n1.y;
			int x2 = (int)n2.x;
			int y2 = (int)n2.y;
			int len = (int)Math.abs(Math.sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2))-e.len);
			offgraphics.setColor(Color.black);
			offgraphics.drawLine(x1, y1, x2, y2);
		}
		
		// draw nodes
		for (Enumeration nodeEnum = nodes.elements(); nodeEnum.hasMoreElements(); ) {
			Node n = (Node)nodeEnum.nextElement();		  
			paintNode(offgraphics, n, fm);
		}
		setSize(sizePanel);
		g.drawImage(offscreen, 0, 0, null);
    }

    // mouse event handling
    public void mouseClicked(MouseEvent e) {
    }

    public void mousePressed(MouseEvent e) {
		addMouseMotionListener(this); 	
		px = e.getX();
		py = e.getY(); 
		pnode = null;
		phost = null;
		
		for (Enumeration nodeEnum = nodes.elements(); nodeEnum.hasMoreElements(); ) {
			Node n = (Node)nodeEnum.nextElement();		
			if (isPointInsideNode(px, py, n)) {
				pnode = n;
				repaint();
				break;
			}
		}
		if (pnode == null && resized == null) {
			for (Enumeration hostEnum = hosts.elements(); hostEnum.hasMoreElements(); ) {
				Host h = (Host)hostEnum.nextElement();
				if (isPointInsideHost(px, py, h)) {
					phost = h;
					repaint();
					break;
				}
			}
		}

		switch (e.getModifiers()) { 	
		case MouseEvent.BUTTON3_MASK:
			Popup popup; 	
			if (pnode != null) 
				popup = new Popup((Graph)getParent().getParent(), Popup.POPUP_PROC); 
			else  		
				popup = new Popup((Graph)getParent().getParent(), Popup.POPUP_ENGINE); 
			add(popup);
			popup.show(this, px, py);
			break;
		default:
			break; 		
		}
		e.consume();
    }
	
    public void mouseReleased(MouseEvent e) {
		if (pnode != null) {
			for (Enumeration hostEnum = hosts.elements(); hostEnum.hasMoreElements(); ) {
				Host h = (Host)hostEnum.nextElement();					
				if (!pnode.host.equals(h.name) 
					&& isPointInsideHost(pnode.x, pnode.y, h)) {
					// the picked node was dragged to a new host
					OkCancelBox box = new OkCancelBox((Frame)getParent().getParent(), 
										 "Dacia", 
										 "Move Proc "+pnode.id+"  to "+h.name+" ?"); 
					if (isOK) {			       
						if (Engine.moveProc(new Integer(pnode.id), h.name)) {
							Engine.updateFlag = true;
							return;
						}
						MessageBox msgBox = new MessageBox((Frame)getParent().getParent(), 
											"Alert", 
											"Cannot move "+pnode.id+" to "+h.name);
					}
					break;			   
				}
			}
	
			// the picked node was not dragged to a new box 	    
			Host h = (Host)hosts.get(pnode.host);
			if (pnode.x-pnode.w/2 < h.x)
				pnode.x = h.x + pnode.w/2;
			else if (pnode.x+pnode.w/2 > h.x+h.w)
				pnode.x = h.x + h.w - pnode.w/2;
			if (pnode.y-pnode.h/2 < h.y)
				pnode.y = h.y + pnode.h/2;
			else if (pnode.y+pnode.h/2 > h.y+h.h)
				pnode.y = h.y + h.h - pnode.h/2;
		}
		e.consume();
    }
    
    public void mouseEntered(MouseEvent e) {
    }

    public void mouseExited(MouseEvent e) {
    }

    public void mouseDragged(MouseEvent e) {
	int px1 = e.getX();
	int py1 = e.getY();

	if (px1 < 0 || py1 < 0)
	    return;

	int dx = px1 - px;
	int dy = py1 - py; 
		
	if (pnode != null) { // moving a node
	    pnode.x += dx;
	    pnode.y += dy; 
	    px = px1;
	    py = py1;
	}
	else if (phost != null) { // moving a box
	    // keep the box from overlapping from others
	    for (Enumeration hostEnum = hosts.elements(); hostEnum.hasMoreElements(); ) {
			Host h = (Host)hostEnum.nextElement(); 	
			if (phost != h) {
				// if the left edge is getting inside another box 	
				if (phost.x >= h.x+h.w
					&& phost.x+dx < h.x+h.w
					&& phost.y+dy < h.y+h.h 			
					&& phost.y+phost.h+dy > h.y) {
					dx = h.x+h.w-phost.x; 		
					px1 = px+dx;
				}
				// if the right edge is getting inside another box 		
				else if (phost.x+phost.w <= h.x
						 && phost.x+phost.w+dx > h.x
						 && phost.y+dy < h.y+h.h 	
						 && phost.y+phost.h+dy > h.y) {
					dx = h.x-phost.w-phost.x; 		
					px1 = px+dx;
				} 			
				// if the top edge is getting inside another box 		
				else if (phost.y >= h.y+h.h
						 && phost.y+dy < h.y+h.h
						 && phost.x+dx < h.x+h.w
						 && phost.x+phost.w+dx > h.x) {
					dy = h.y+h.h-phost.y; 			
					py1 = py+dy; 			
				}
				// if the bottom edge is getting inside another box 	
				else if (phost.y+phost.h <= h.y
						 && phost.y+phost.h+dy > h.y
						 && phost.x+dx < h.x+h.w
						 && phost.x+phost.w+dx > h.x) {
					dy = h.y-phost.h-phost.y; 			
					py1 = py+dy; 		
				}
			}
	    }
	    for (Enumeration nodeEnum = nodes.elements(); nodeEnum.hasMoreElements(); ) {
			Node n = (Node)nodeEnum.nextElement();		
			if (n.host.equals(phost.name)) {
				n.x += dx;
				n.y += dy;
			}
	    }
	    phost.x += dx;
	    phost.y += dy; 	
	    px = px1;
	    py = py1;
	    getMinSize();
	}
	else if (resized != null) { // resizing a box
	    // the box size cannot be smaller than MIN_BOX_WIDTH x MIN_BOX_HEIGHT
	    if (resized.w+dx >= MIN_BOX_WIDTH) {
			resized.w += dx;
			px = px1;
	    }
	    if (resized.h+dy >= MIN_BOX_HEIGHT) {
			resized.h += dy;
			py = py1;
	    }
	    // keep the box from overlapping from others
	    for (Enumeration hostEnum = hosts.elements(); hostEnum.hasMoreElements(); ) {
			Host h = (Host)hostEnum.nextElement();
			if (resized != h) {
				// if the right edge is getting inside another box 		
				if (resized.x+resized.w-dx <= h.x
					&& resized.x+resized.w > h.x
					&& resized.y < h.y+h.h 			
					&& resized.y+resized.h > h.y) { 		
					resized.w = h.x - resized.x;
					px = h.x;
				}
				// if the bottom edge is getting inside another box 		
				else if (resized.y+resized.h-dy <= h.y
						 && resized.y+resized.h > h.y
						 && resized.x < h.x+h.w
						 && resized.x+resized.w > h.x) { 		
					resized.h = h.y - resized.y; 		
					py = h.y; 
				}
			}
	    }
	    // if nodes are outside the box, move them together 	
	    for (Enumeration nodeEnum = nodes.elements(); nodeEnum.hasMoreElements(); ) {
			Node n = (Node)nodeEnum.nextElement();	
			Host h = (Host)hosts.get(n.host);
			if (resized == h) { 
				if (n.x + n.w/2 > resized.x + resized.w) 	
					n.x = resized.x + resized.w - n.w/2;	
				if (n.y + n.h/2 > resized.y + resized.h)  
					n.y = resized.y + resized.h - n.h/2;	
			}
	    }		 	   
	    getMinSize();
	}		
	e.consume();  
    }

    public void mouseMoved(MouseEvent e) {
		int x = e.getX();
		int y = e.getY();		
		for (Enumeration hostEnum = hosts.elements(); hostEnum.hasMoreElements(); ) {
			Host h = (Host)hostEnum.nextElement();
			if (x>=h.x+h.w-TOLERANCE && x<=h.x+h.w+TOLERANCE && 
				y>=h.y+h.h-TOLERANCE && y<=h.y+h.h+TOLERANCE) {
				resized = h;
				setCursor(Cursor.getPredefinedCursor(Cursor.SE_RESIZE_CURSOR));
				//repaint();
				e.consume();  
				return;
			}
		}
		setCursor(Cursor.getDefaultCursor()); 
		resized = null;
    }

    public void start() {
		relaxer = new Thread(this);
		relaxer.start();
    }

    public void stop() {
		relaxer = null;
    }
	
}


/**
 * The information panel is the area for displaying text information. The 
 * information displayed is the same as using "print" in the shell interface.
 */
class InfoPanel extends Panel implements ActionListener {

    TextArea info1;
    TextArea info2;
     	
    public InfoPanel() { 
    setFont(new Font("sanserif", Font.PLAIN, 12));
	GridBagLayout g = new GridBagLayout();
	setLayout(g); 			 
	GridBagConstraints gbc = new GridBagConstraints(); 		
     	
	// TextArea for connection info
	info1 = new TextArea(Engine.connectionsInfo(), 6, 40);
	info1.setEditable(false); 	
	//info1.setBackground(getBackground());
	gbc.insets = new Insets(10, 10, 0, 10);
	gbc.weightx = 1.0; 
	gbc.weighty = 1.0; 	
	gbc.gridx = 1; 	
	gbc.gridy = 1;
	gbc.gridheight = 1; 
	gbc.fill = GridBagConstraints.BOTH;
	gbc.anchor = GridBagConstraints.WEST;
	g.setConstraints(info1, gbc); 
	add(info1);
	
	// TextArea for proc info
	info2 = new TextArea(Engine.procsInfo(), 14,40);
	info2.setEditable(false);
	info2.setBackground(getBackground());
	gbc.gridy = 2;
	g.setConstraints(info2, gbc); 
	add(info2);
	
	// Close and Update buttons
	Panel p = new Panel();
	Button close; 
	close = new Button("Close");
	close.addActionListener(this);
	p.add(close);
	gbc.weightx = 0; 
	gbc.weighty = 0; 
	gbc.gridy = 3;
	gbc.anchor = GridBagConstraints.CENTER;
	g.setConstraints(p, gbc);
	add(p); 
	
	setSize(280, 400);
    }
    
    public void engineUpdate() {
		info1.setText(Engine.connectionsInfo());
		info2.setText(Engine.procsInfo());
    }
       
    public void actionPerformed(ActionEvent e) {
		String label = ((Button)e.getSource()).getLabel(); 		
		if (label.equals("Close")) {
			Graph graph = (Graph)getParent();
			graph.remove(this);
			graph.pack();
		}	
     }
			
}  


/**
 * Graph is the graphical interface for DACIA. It has all the functions 
 * provided by the shell interface. It is an optional tool that allows user 
 * to visualize the application components and interactively reconfigure 
 * those components.
 * @see Engine
 * @author Yizhou Cao
 */

public class Graph extends Frame 
	implements Runnable, ActionListener {

	/** The graph panel */
    GraphPanel  graphPanel; 	
	/** The information panel */
    InfoPanel	infoPanel;
    private ScrollPane  scrollPane;

    Thread updateThread;
    private static int INTERVAL = 1000;

    Graph(String title) {
		super(title);

		MenuBar 	menuBar;  
		Menu      menu;
		MenuItem	m;

		menuBar = new MenuBar();
		setMenuBar(menuBar);	    
	 
		// menu "File"
		menu = new Menu("File");
		m = new MenuItem("Close"); 
		m.addActionListener(this); 
		menu.add(m);
		menu.addSeparator();
		m = new MenuItem("Quit"); 
		m.addActionListener(this); 
		menu.add(m);	
		menuBar.add(menu);
		
		// menu "View"
		menu = new Menu("View");
		m = new MenuItem("Redraw"); 
		m.addActionListener(this); 
		menu.add(m);
		m = new MenuItem("Show Info");
		m.addActionListener(this); 
		menu.add(m);
		m = new MenuItem("Preferences"); 
		m.addActionListener(this); 
		menu.add(m);
		menuBar.add(menu);
		
		// menu "Engines"
		menu = new Menu("Engines"); 
		m = new MenuItem("Connect Engines");
		m.addActionListener(this); 
		menu.add(m); 
		m = new MenuItem("Start Monitor");
		m.addActionListener(this); 
		menu.add(m); 
		m = new MenuItem("Stop Monitor");
		m.addActionListener(this); 	
		menu.add(m); 	
		m = new MenuItem("Update host");
		m.addActionListener(this); 
		menu.add(m); 	
		//menu.add(new MenuItem("-")); 
		menu.addSeparator();
		m = new MenuItem("Load class");
		m.addActionListener(this); 
		menu.add(m); 	
		menuBar.add(menu);
		
		// menu "Procs"
		menu = new Menu("Procs");
		m = new MenuItem("Start Proc");
		m.addActionListener(this); 
		menu.add(m); 
		m = new MenuItem("Connect Procs");
		m.addActionListener(this); 	
		menu.add(m); 	
		m = new MenuItem("Disconnect Proc");
		m.addActionListener(this); 	
		menu.add(m); 
		m = new MenuItem("Move Proc");
		m.addActionListener(this); 
		menu.add(m); 
		m = new MenuItem("Remove Proc");
		m.addActionListener(this); 
		menu.add(m); 
		menuBar.add(menu); 
	 	
		// menu "Help"
		menu = new Menu("Help");
		m = new MenuItem("About"); 
		m.addActionListener(this); 
		menu.add(m);	
		menuBar.add(menu);
		
		setLayout(new BorderLayout());
		
		// add ScrollPanel
		scrollPane = new ScrollPane (ScrollPane.SCROLLBARS_AS_NEEDED);
		add("Center", scrollPane);

		// add GraphPanel
		graphPanel = new GraphPanel();
		scrollPane.add(graphPanel);
		graphPanel.start(); 	
		graphPanel.sizeVP = scrollPane.getViewportSize();
		
		// add InfoPanel
		infoPanel = new InfoPanel();	
		add("East", infoPanel);
		
		// add listener for size change of the ScrollPane
		scrollPane.addComponentListener(new ComponentAdapter() {
			public void componentResized(ComponentEvent e) {
				graphPanel.offscreen = null;
			}
		} );
		
		// Handle window close requests
		addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				graphPanel.stop();
				removeAll();
				Engine.graph = null;
				dispose(); 
			}
		});

		setSize(900, 500); 
		graphPanel.setSize(scrollPane.getViewportSize());

		start();
    }

    public void run() {
      while (true) {
        if (Engine.updateFlag)
          updateGraph();
        try {
          Thread.sleep(INTERVAL);
        } 
        catch (InterruptedException e) {
          break;
        } 
      }
    }

    public void start() {
		updateThread = new Thread(this);
		updateThread.start();
    }

    public void stop() {
		updateThread = null;
    }
	
	public void updateGraph() {
		if (graphPanel != null)
			graphPanel.engineUpdate();
		if (infoPanel != null)
			infoPanel.engineUpdate();
		Engine.updateFlag = false;
	}

	public void redrawAll() {
		if (graphPanel != null)
			graphPanel.redrawAll();
	}

	public void removeHost(String name) {
		if (graphPanel != null)
			graphPanel.removeHost(name);
	}

	public void removeNode(int id) {
		if (graphPanel != null)
			graphPanel.removeNode(id);
	}

  public void actionPerformed(ActionEvent e) {
		String cmd = e.getActionCommand();
		if (cmd.equals("Quit"))  	
			System.exit(0);
		if (cmd.equals("Close")) {
			graphPanel.stop();
			stop();
			removeAll();
			//setVisible(false);
			Engine.graph = null;
			dispose(); 
		}
		else if (cmd.equals("Redraw")) {
			graphPanel.redrawAll();
			infoPanel.engineUpdate();
		}
		else if (cmd.equals("Show Info")) {
			if (infoPanel == null)  		
				infoPanel = new InfoPanel(); 	
			add("East", infoPanel); 	
			pack(); 
		}
		else if (cmd.equals("Start Proc")) {
			StartProcDialog spdlg = new StartProcDialog(this); 
		} 
		else if (cmd.equals("Preferences")) {
			PreferDialog preferDlg = new PreferDialog(this);
		}		
		else if (cmd.equals("Connect Engines")) {
			ConnectEngineDialog engineDlg = new ConnectEngineDialog(this); 
		}
		else if (cmd.equals("Start Monitor")) 
			Engine.startMonitor(); 
		else if (cmd.equals("Stop Monitor")) 
			Engine.stopMonitor(); 
		else if (cmd.equals("Update host")) {
			UpdateDialog updateDlg = new UpdateDialog(this);
		} 
		else if (cmd.equals("Load class")) { 
			LoadClassDialog loaderDlg = new LoadClassDialog(this);
		} 
		else if (cmd.equals("Connect Procs")) { 
			ConnectProcsDialog connectDlg = new ConnectProcsDialog(this); 
		} 	 
		else if (cmd.equals("Disconnect Proc"))  { 		 
			DisconnectProcDialog disconnectDlg = new DisconnectProcDialog(this); 	 
	    } 	 
		else if (cmd.equals("Move Proc")) {	 
			MoveProcDialog moveDlg = new MoveProcDialog(this); 
		} 	 
		else if (cmd.equals("Remove Proc")) {	 
			RemoveProcDialog removeDlg = new RemoveProcDialog(this); 
		} 	 
		else if (cmd.equals("About"))  { 
			MessageBox about = new MessageBox((Frame)this, "About",   		
					      "Dacia\n"+
                          "by Radu Litiu (radu@eecs.umich.edu)\n"+
                          "   Yizhou Cao (caoy@umich.edu)\n"+
                          "Copyright (c) 2000, U of M");		
    } 	 
  }
	
}


// Some utility methods

class Util {
    
    public static void getUnusedPorts (Choice proc, Choice port) {
		Proc       lproc;
		RemoteProc rproc;
		
		int procID = Integer.parseInt(proc.getSelectedItem());
		port.removeAll();
		if ((lproc = (Proc)Engine.getLocalProcs().get(new Integer(procID))) != null) {
			for (int i = 0 ; i < lproc.portNo; i++) {
				if (lproc.port[i].connectionStatus == 0)
					port.addItem(String.valueOf(i));
			}
			if (port.getItemCount() == 0)
				port.addItem(" ");
		}
		else if ((rproc = (RemoteProc)Engine.getRemoteProcs().get(new Integer(procID)))
				 != null) {
			for (int i = 0 ; i < rproc.portNo; i++) { 
				if (rproc.connectedProcID[i] == 0)
					port.addItem(String.valueOf(i)); 
			}
		}
		
		if (port.getItemCount() == 0)
			port.addItem(" ");	
    }

    public static void getUsedPorts (Choice proc, Choice port) {
		Proc       lproc;
		RemoteProc rproc;
		
		int procID = Integer.parseInt(proc.getSelectedItem());
		port.removeAll();
		if ((lproc = (Proc)Engine.getLocalProcs().get(new Integer(procID))) != null) {
			for (int i = 0 ; i < lproc.portNo; i++) {
				if (lproc.port[i].connectionStatus != 0)
					port.addItem(String.valueOf(i));
			}
		}
		else if ((rproc = (RemoteProc)Engine.getRemoteProcs().get(new Integer(procID)))
				 != null) {
			for (int i = 0 ; i < rproc.portNo; i++) { 
				if (rproc.connectedProcID[i] != 0)
					port.addItem(String.valueOf(i)); 
			}
		}
		
		if (port.getItemCount() == 0)
			port.addItem(" ");
    }

}
