package dacia;

/* Implements the MERGE PROC.
 * By default, it has 2 input (ports 0 and 1) and 1 output (port 2) ports.
 * It merges two by two the messages received on the input ports, and sends
 * the result to the output port.
 * The abstract merge() method have to be defined in the subclass.
 */
public abstract class Merge extends Proc {
  public Merge(String procName) {
		super(procName, 3);
		type = MERGE;
	}

	/* A Merge PROC should handle a message asynchronously. A message received
   * synchronously on a port is re-sent asynchronously to the same port.
   */
	public void handleMessage(Message msg, int portNo) {
    asyncMessageReceived(msg, portNo);
	}

  /* Waits until a message is received on both ports.
   * Merges the two messages and outputs the result.
   */
	public void handleAsyncMessage() {
	  Message m1 = null, m2 = null;

    while(true) {
      synchronized(msgQueue) {
        while((msgNo[0]*msgNo[1] == 0) && !moving) {
          //System.out.println("Merge waiting");
          try {
            msgQueue.wait();
          }
          catch (Exception e) {
            System.out.println("Merge: exception raised in wait(): "
                               + e.getMessage());
            e.printStackTrace();
          }
        }
        if(moving) {
          return;
        }
        m1 = getMessage(0, true);
        m2 = getMessage(1, true);
      }
      m1 = merge(m1, m2);
      output(2, m1, 1);
    }
	}

  /* This abstract method should be defined in the subclass.
	 * It merges two messages into a single message.
   */
	abstract Message merge(Message m1, Message m2);
}

/* Old syntax.
 * Copies the input for both input ports to the output port.
 * For more complex functionality, it has to be subclassed and the
 * handleMessage() methods have to be modified.
 */
