package dacia;

import java.io.Serializable;

/* Implements a port, used to interconnect two components.
 * The same port works both as input and output port.
 * @see Proc
 * @see Engine
 * @author Radu Litiu
 */
public class Port implements Serializable {
  Proc myProc;
  /* The connection status for this port
   * 0 - disconnected;
   * 1 - connected locally
   * 2 - connected remotely
   */
  public int connectionStatus = 0;
  /* Specifies whether this port is used primarily as source for a connection.
   * This information may be useful in building a directed graph
   * corresponding to an application.
   */
  boolean source = false;
  /* Identifies the connected port, if there is one. A connection can be
   * local (connectedProc + connectedProcID + connectedPortNo) or
   * remote (remoteConnection + connectedProcID + connectedportNo).
   */
  public Proc connectedProc = null;
  public int connectedProcID = 0;
  public int connectedPortNo = 0;
  Connection remoteConnection = null;
  /* the output type
   * 1 - synchronous, 0 - asynchronous communication
   */
  private int outputType = 1;

  protected Port(Proc proc) {
    myProc = proc;
  }

  protected Port(Proc proc, int synchronous) {
    myProc = proc;
    outputType = synchronous;
  }

  /* Connects to a local port
   */
  void connect(Proc proc, int portNo, boolean isSource) {
    connectionStatus = 1;
    connectedProc = proc;
    connectedProcID = proc.id;
    connectedPortNo = portNo;
    source = isSource;
  }

  /* Connects to a remote port
   */
  void connect(int procID,int portNo,Connection connection,boolean isSource) {
    connectionStatus = 2;
    connectedProcID = procID;
    connectedPortNo = portNo;
    remoteConnection = connection;
    source = isSource;
  }

  /* Disconnects this port
   */
  void disconnect() {
    connectionStatus = 0;
    connectedProc = null;
    connectedProcID = 0;
    connectedPortNo = 0;
    remoteConnection = null;
  }

  void setOutputType(int type) {
    outputType = type;
  }

  int getOutputType() {
    return outputType;
  }

  public Connection getRemoteConnection() {
    return remoteConnection;
  }

}
