package dacia;

import java.io.Serializable;
import java.io.*;

/* The local image of a remote Proc
 * @see Proc
 * @see Engine
 * @author Radu Litiu
 */
public class RemoteProc implements Serializable {
  int id;
  int type;
  public String name;
  public String hostname;
  // The connection to the Engine where this RemoteProc resides
  Connection connection = null;
  /* Identify the connections
   * connectedProcID represents the IDs of the PROCs connected to these
   * ports. connectedPortNo represents the port #s of the PROCs connected
   * to these ports.
   */
  public int portNo, connectedProcID[], connectedPortNo[];

  /* copy constructor; initializes the connection to null
   */
  public RemoteProc(RemoteProc proc) {
    id = proc.id;
    type = proc.type;
    name = proc.name;
    hostname = proc.hostname;
    portNo = proc.portNo;
    connectedProcID = new int[portNo];
    connectedPortNo = new int[portNo];
    for(int i=0; i<portNo; i++) {
      connectedProcID[i] = proc.connectedProcID[i];
      connectedPortNo[i] = proc.connectedPortNo[i];
    }
  }

  public RemoteProc(Proc proc) {
    id = proc.id;
    type = proc.type;
    name = proc.name;
    hostname = proc.hostname;
    portNo = proc.portNo;
    connectedProcID = new int[portNo];
    connectedPortNo = new int[portNo];
    for(int i=0; i<portNo; i++) {
      if(proc.port[i].connectionStatus != 0) {
        connectedProcID[i] = proc.port[i].connectedProcID;
        connectedPortNo[i] = proc.port[i].connectedPortNo;
      }
      else {
        connectedProcID[i] = 0;
        connectedPortNo[i] = 0;
      }
    }
  }

  /* Prints information about this remote Proc and its connections.
   */
  public void printInfo() {
    System.out.println("    " + id + " " + name + ":" + type + ", " +
                       portNo + " ports, host: " + hostname);
    for(int i=0; i<portNo; i++) {
      if(connectedProcID[i] !=0) {
        System.out.println("\t" + i + " -> " + connectedProcID[i] + ":" +
                           connectedPortNo[i]);
      }
    }
  }

  /* Prints information about this remote Proc and its connections
   * to the indicated output stream.
   * @param dos  the output stream to print the info to
   */
  public void printInfo(DataOutputStream dos) {
    try{
      dos.writeBytes("    " + id + " " + name + ":" + type + ", " +
                     portNo + " ports, host: " + hostname + "\n");
      for(int i=0; i<portNo; i++)
        if(connectedProcID[i] !=0)
          dos.writeBytes("\t" + i + " -> " + connectedProcID[i] + ":" +
                         connectedPortNo[i] + "\n");
    }
    catch(IOException e) {
      System.out.println("Engine: exception in printInfo: " + e.getMessage());
      e.printStackTrace();
    }
  }

  public String getName() {
    return name;
  }

  public int getID() {
    return id;
  }

  public Connection getConnection() {
    return connection;
  }
}
