package dacia;

/* Implements the TRANSFORM PROC, having 1 input (port 0) and 1 output
 * (port 1) ports. Applies a transformation to the messages received on
 * the input port and outputs the results to the output port.
 */
public abstract class Transform extends Proc {
  public Transform(String procName) {
		super(procName, 2);
		type = TRANSFORM;
	}

	/* handleMessage() and handleAsyncMessage() should do the same thing,
   * call handleMessage(msg), which should be defined in the subclass
   */
	public void handleMessage(Message msg, int port) {
    if(port == 1)
      System.out.println("Transform " + id + " : message received on port 1");
    else
      handleMessage(msg);
	}

	public void handleAsyncMessage() {
		Message msg = null;
    while(true) {
      if((msg = getMessage()) == null)
        return;
      if(msg.getPort() == 1)
        System.out.println("Transform " + id +
                           " : async message received on port 1");
      else
        handleMessage(msg);
    }
	}

  /* This abstract method should be defined in the subclass.
   * The message resulted should be sent on port 1.
   */
	public abstract void handleMessage(Message msg);
}


