package dacia.apps.chat;

import dacia.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.Font;
import java.util.*;
import java.net.*;
import java.io.*;

/* Implements the CHAT PROC, having 1 port. Displays in the output textArea
 * the messages received on the port. Outputs the messages typed in the
 * input textArea.
 */
public class Chat extends Proc {
  ChatFrame frame = null;
  String text = null;
  int msgID = 0;
  boolean isParked = false;
  String mailTo = null;

  public Chat() {
    super("Chat", 1);
    frame = new ChatFrame(this);
    //frame.init();
  }

  public void handleMessage(Message msg, int port) {
    synchronized(this) {
      if(!isParked)
        frame.displayText(new String(msg.origin + ": " + (String)msg.getData()
                                     + "\n"));
      else {
        text = new String(text +msg.origin +": " +(String)msg.getData() +"\n");
        if(mailTo != null) {
          // send email to the specified address
          System.out.println("Send mail to " + mailTo + ": " +
                             (String)msg.getData());
          /*
          try{
            Runtime.getRuntime().exec("mail -s DACIA_msg radu@eecs.umich.edu\n mail1\n\n.\nw");
          }
          catch(Exception e) {
            System.out.println("mail error: " + e.getMessage());
            e.printStackTrace();
          }
          */
          sendmail(mailTo, msg.origin, (String)msg.getData());
        }
      }
    }
  }

	public void handleAsyncMessage() {
		Message msg = null;
    while(true) {
      if((msg = getMessage()) == null)
        return;
      handleMessage(msg, 0);
    }
	}

  void sendMessage(String message) {
    Message msg = new Message();
    msg.id = msgID++;
    msg.origin = getID();
    msg.setData((Object)message);
    output(0, msg, 1);
  }

  public void pack() {
    if(!isParked) {
      text = frame.getText();
      frame.quit();
      frame = null;
    }
  }

  public void unpack() {
    if(!isParked) {
      frame = new ChatFrame(this);
      frame.displayText(text);
    }
  }

  void park(String hostname, String emailaddr) {
    System.out.println("park(" +getID() +", " +hostname +", " + emailaddr+")");
    mailTo = emailaddr;
    synchronized(this) {
      pack();
      isParked = true;
    }
    if(hostname != null) {
      Connection connection;
      if((connection = Engine.getConnection(hostname)) != null) {
        Engine.moveProc(this, connection);
      }
    }
  }

  public void start() {
    if(isParked) {
      synchronized(this) {
        isParked = false;
        unpack();
      }
      mailTo = null;
    }
  }

  void sendmail(String recipient, int sender, String message) {
    String hostname = null;

    int i = recipient.indexOf("@");
    if(i > 0) {
      hostname = recipient.substring(i+1);
    }
    try {   
      //Socket s = new Socket(mailServer, 25);
      Socket s = new Socket(hostname, 25);

      BufferedReader in = new BufferedReader
        (new InputStreamReader(s.getInputStream(), "8859_1"));
      BufferedWriter out = new BufferedWriter
        (new OutputStreamWriter(s.getOutputStream(), "8859_1"));

      // warning : some mail server validate the sender address
      //           in the MAIL FROM command, put your real address here
      sendln(in, out, "MAIL FROM: <radu@eecs.umich.edu>");
      // sendln(in, out, "RCPT TO: <radu@eecs.umich.edu>" );
      sendln(in, out, "RCPT TO: <" + recipient + ">");
      sendln(in, out, "DATA");
      sendln(out, "Subject: DACIA chat message");
      sendln(out, "From: Chat " + (new Integer(sender)).toString() +
             " <chat@DACIA>");
      //sendln(out, "To: <radu@eecs.umich.edu>");
      sendln(out, "To: <" + recipient + ">");

      // Send the body
      sendln(out, message);
      sendln(in, out,".");
      sendln(in, out, "QUIT");
      s.close();
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }
   
  public void sendln(BufferedReader in, BufferedWriter out, String s) {
    try {
      out.write(s + "\r\n");
      out.flush();
      // System.out.println(s);
      s = in.readLine();
      // System.out.println(s);
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }

  public void sendln(BufferedWriter out, String s) {
    try {
      out.write(s + "\r\n");
      out.flush();
      System.out.println(s);
    }  
    catch (Exception e) {
      e.printStackTrace();
    }
  } 

}


class ChatFrame extends Frame implements ActionListener{
  Chat chat;
  Button sendButton = null;
  MenuItem startItem = null;
  MenuItem stopItem = null;
  MenuItem parkItem = null;
  TextField editText = null;
  TextArea displayArea = null;

  public ChatFrame(Chat target) {
    super("Chat: " + target.getID());
    chat = target;
    setSize(350, 250);
    Font font = new Font("sanserif", Font.PLAIN|Font.BOLD, 14);
    setFont(font);

    // create the user interface
    GridBagLayout gridbag = new GridBagLayout();
    setLayout(gridbag);

    sendButton = new Button("Send");
    editText = new TextField();
    displayArea = new TextArea(7,40);
    displayArea.setEditable(false);

    constrain(this, new Label("Enter a new message:"), 0, 0, 2, 1, 0, 0, 0, 0);
    constrain(this, editText, 0, 1, 1, 1, 
		   GridBagConstraints.HORIZONTAL,
		   GridBagConstraints.NORTHWEST,
		   10.0, 0.0, 0, 0, 0, 0);
    constrain(this, sendButton, 1, 1, 1, 1, 
		   GridBagConstraints.HORIZONTAL,
		   GridBagConstraints.NORTHWEST,
		   1, 0.0, 0, 0, 0, 0);
    constrain(this, new Label("Latest message:"), 0, 2, 2, 1, 
		   GridBagConstraints.HORIZONTAL,
		   GridBagConstraints.NORTHWEST,
		   1.0, 0.0, 0, 0, 0, 0);
    constrain(this, displayArea, 0, 3, 2, 1, 
		   GridBagConstraints.BOTH,
		   GridBagConstraints.CENTER,
		   1.0, 1.0, 0, 0, 1, 0);

    // set up the menu bar
    Menu aMenu = new Menu("Menu");
    startItem = new MenuItem("Start");
    startItem.addActionListener((ActionListener)this);
    aMenu.add(startItem);
    parkItem = new MenuItem("Park");
    parkItem.addActionListener((ActionListener)this);
    aMenu.add(parkItem);
    stopItem = new MenuItem("Quit");
    stopItem.addActionListener((ActionListener)this);
    aMenu.add(stopItem);
    MenuBar aMenuBar = new MenuBar();
    aMenuBar.add(aMenu);
    setMenuBar(aMenuBar);
    sendButton.addActionListener(this);
    show();

  }

  //process menu commands
  public void actionPerformed(ActionEvent e) {
    try{
      if(e.getActionCommand().equals("Start")) {
      }
      else if(e.getActionCommand().equals("Park")) {
        ParkDialog dialog = new ParkDialog(this);
      }
      else if(e.getActionCommand().equals("Quit")) {
        setVisible(false);
        dispose();
        //System.exit(0);
        //Thread.currentThread().stop();
      }
      else if(e.getActionCommand().equals("Send")) {
        displayArea.append("OUT : " + editText.getText() + "\n");
        chat.sendMessage(editText.getText());
        editText.setText("");
        editText.requestFocus();
      }
      else
        System.out.println("Chat: unknown menu item: " + e.getActionCommand());
    }
    catch(Exception ex) {
      System.out.println("Chat: exception: "+ex.getMessage());
      ex.printStackTrace();
    }
  }

  void quit() {
    setVisible(false);
    dispose();
  }

  // the gridbag layout stuff from the Java NutShell book
  void constrain(Container container, Component component,
			int grid_x, int grid_y, int grid_width, int grid_height,
			int fill, int anchor, double weight_x, double weight_y,
			int top, int left, int bottom, int right) {
    GridBagConstraints c = new GridBagConstraints();
    c.gridx = grid_x; c.gridy = grid_y;
    c.gridwidth = grid_width; c.gridheight = grid_height;
    c.fill = fill; c.anchor = anchor;
    c.weightx = weight_x; c.weighty = weight_y;
    if (top+bottom+left+right > 0)
      c.insets = new Insets(top, left, bottom, right);
    
    ((GridBagLayout)container.getLayout()).setConstraints(component, c);
    container.add(component);
  }

  void constrain(Container container, Component component,
			int grid_x, int grid_y, int grid_width, int grid_height,
			int top, int left, int bottom, int right) {
    constrain(container, component, grid_x, grid_y,
              grid_width, grid_height, GridBagConstraints.NONE,
              GridBagConstraints.NORTHWEST,
              0.0, 0.0, top, left, bottom, right);
  }

  String getText() {
    return displayArea.getText();
  }

  void displayText(String str) {
    displayArea.append(str);
  }

  void displayMessage(String str) {
    displayArea.append("\nIN : " + str);
  }
}


class ParkDialog extends Dialog implements ActionListener { 
  ChatFrame chatFrame;
	Checkbox move, mail;
  Choice host; 	
  TextField email; 	

  public ParkDialog(ChatFrame parent) { 
		super(parent, "Park Chat client", true);
		chatFrame = parent;
		setFont(new Font("sanserif", Font.PLAIN, 12));
		
		GridBagLayout g = new GridBagLayout();
		setLayout(g);	
		GridBagConstraints gbc = new GridBagConstraints(); 

		move = new Checkbox("Move to host", false);
		gbc.weighty = 1.0; 	
		gbc.insets = new Insets(0, 10, 0, 0);
		gbc.gridx = 1; 	
		gbc.gridy = 1;	
		gbc.gridwidth = 1;  		
		g.setConstraints(move, gbc); 
		add(move);

		host = new Choice();
		gbc.gridx = 2; 	
		gbc.gridwidth = 1; 
		g.setConstraints(host, gbc); 
		add(host);
		
		mail = new Checkbox("Send email to", false);
		gbc.insets = new Insets(0, 10, 0, 0);
		gbc.gridx = 1; 	
		gbc.gridy = 2;	
		gbc.gridwidth = 1;  		
		g.setConstraints(mail, gbc); 
		add(mail);

		email = new TextField(30); 		
		gbc.gridx = 2; 		
		gbc.gridwidth = 2; 
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(email, gbc); 
		add(email); 

		Button ok = new Button("OK");  	
		ok.addActionListener(this);
		gbc.gridx = 2; 		
		gbc.gridy = 4;
		//gbc.anchor = GridBagConstraints.EAST;
		g.setConstraints(ok, gbc); 
		add(ok);  		
		
		Button cancel = new Button("Cancel");  		
		cancel.addActionListener(this);
		gbc.gridx = 3; 	
		//gbc.anchor = GridBagConstraints.CENTER;
		g.setConstraints(cancel, gbc); 
		add(cancel);  	 

    host.addItem("          ");
    Vector connections = Engine.getConnections();
    for(int i=0; i<connections.size(); i++)
      host.addItem(((Connection)connections.elementAt(i)).getHostname());

		setSize(370, 150);
		setResizable(false); 
		show();
	} 
	
	public void actionPerformed(ActionEvent e) {
    String hostname = null, emailaddr = null;
    //String hostname = "hostname", emailaddr = "emailaddr";
		String label = ((Button)e.getSource()).getLabel();
		if(label.equals("OK")) {
			if(move.getState())
        if(!host.getSelectedItem().equals("          "))
          hostname = host.getSelectedItem();
			if(mail.getState())
				emailaddr = email.getText();
			dispose();
      chatFrame.chat.park(hostname, emailaddr);
		}
		if (label.equals("Cancel"))
			dispose(); 
	}
} 


