package dacia.apps.chat;

import dacia.*;
import java.util.*;
import java.lang.Math;

/* Implements a monitor for the multi-user Chat application.
 * It balances the load across multiple servers and distributes clients
 * accordingly.
 */
class ChatMonitor extends Monitor {
  int MAX_SERVERS = 10;
  int MAX = 6;
  int MIN = 4;
  int INTERVAL = 5000;

  Vector servers = null;

  int nhosts = 3;
  String hosts[] = {"saturn.eecs.umich.edu", "sanjuan.eecs.umich.edu", "seoul.eecs.umich.edu"};

  public void run() {
    Enumeration procs = null;
    Proc localServer = null;
    RemoteProc remoteServer = null;
    int cnt, i, j;
    Node n1, n2;
    boolean success;
    System.out.println("ChatMonitor started.");
    while(true) {
      try {
        Thread.currentThread().sleep(INTERVAL);
      }
      catch(InterruptedException e) {
        System.out.println("Monitor interrupted while sleeping");
      }
      servers = procSelect("ChatServer", 1);
      // balance overloaded servers
      // find the server with the highest load
      if((i = (n1 = (Node)servers.elementAt(0)).connections - MAX + 1) > 0) {
        // server is overloaded
        // find the server with the lowest load
        if((j = MAX - 1 - (n2=(Node)servers.elementAt(servers.size() -
                                                      1)).connections) > 0) {
          // can offload some clients from n1 to n2 - determine their number
          if(i < j)
            cnt = i;
          else
            cnt = j;
          // move cnt PROCs from n1 to n2
          transfer(n1, n2, cnt);
        }
        else { // need to spawn a new server
          success = false;
          String otherHost = null;
          for(i=0; i<nhosts; i++) {
            if((j = hosts[i].indexOf(".")) < 0)
              otherHost = hosts[i];
            else
              otherHost = hosts[i].substring(0, j);

            if(!(otherHost.equals(hostname)) &&
               (Engine.getConnection(otherHost) == null)) {
              Engine.connect(hosts[i], Engine.getPort(), true);
              // wait for the updates from the remote host
              try {
                Thread.currentThread().sleep(500);
              }
              catch(InterruptedException e) { }
              procs = remoteProcs.elements();
              while(procs.hasMoreElements()) {
                if(((remoteServer = (RemoteProc)procs.nextElement()).hostname.equals(otherHost)) && remoteServer.getName().equals("ChatServer")) {
                  System.out.println("Found ChatServer " + remoteServer.getID());
                  // connect the new server with the local server
                  procs = localProcs.elements();
                  while(procs.hasMoreElements())
                    if((localServer = (Proc)(procs.nextElement())).getName().equals("ChatServer")) {
                      Engine.connectProcs(remoteServer,0,localServer,-1,true);
                      break;
                    }
                  if(n1.proc != null)
                    transfer(n1.proc, remoteServer, n1.connections/2);
                  else if(n1.rp != null)
                    transfer(n1.rp, remoteServer, n1.connections/2);
                  success = true;
                  break;
                }
                else
                  System.out.println("RP is not server " + remoteServer.getID() + " " + remoteServer.getName());
              }
              if(success == true) break;
            }
          }
        }
      }
      // balance underloaded servers
      if(((n1 =(Node)servers.elementAt(servers.size()-1)).connections < MIN) &&
         ((n2 =(Node)servers.elementAt(servers.size()-2)).connections < MIN)) {
        // 2 servers are underloaded
        if(n1.rp != null) { // n1 corresponds to a remote host
          // transfer all the clients from one server to the other one
          transfer(n1, n2, n1.connections);
          // allow for outstanding structural changes to complete
          try {
            Thread.currentThread().sleep(500);
          }
          catch(Exception e) { }
          // deallocate the free server
          System.out.println("Disconnecting host " + n1.rp.hostname);
          Engine.disconnectHost(n1.rp.getConnection());
        }
      }
    }
  }

  /* Returns an array containing the PROCs of the type indicated,
   * sorted in decreasing order of their number of connections.
   * @param procName  the name of the PROCs sought
   * @param localORremote  indicates whether to return the local or the remote PROCs: 0 - local only; 1 - all PROCs; 2 - remote only
   */
  Vector procSelect(String procName, int localORremote) {
    Proc proc;
    RemoteProc rp;
    Node node;
    Vector proclist = new Vector(MAX_SERVERS);
    int i, size;
    Enumeration procs;
    if(localORremote <= 1) {
      // add the local PROCs
      procs = localProcs.elements();
      while(procs.hasMoreElements()) {
        if((proc = (Proc)procs.nextElement()).getName().equals(procName)) {
          node = new Node(proc);
          size = proclist.size();
          for(i=0; i<size; i++)
            if(node.connections > ((Node)proclist.elementAt(i)).connections) {
               proclist.insertElementAt(node, i);
               break;
            }
          if(i == size)
            proclist.addElement(node);
        }
      }
    }
    if(localORremote >= 1) {
      // add the remote PROCs
      procs = remoteProcs.elements();
      while(procs.hasMoreElements()) {
        if((rp = (RemoteProc)procs.nextElement()).getName().equals(procName)) {
          node = new Node(rp);
          size = proclist.size();
          for(i=0; i<size; i++)
            if(node.connections > ((Node)proclist.elementAt(i)).connections) {
              proclist.insertElementAt(node, i);
              break;
            }
          if(i == size)
            proclist.addElement(node);
        }
      }
    }

    Node n;
    // System.out.println("\nprocSelect(" + procName + ")");
    System.out.println("");
    for(i=0; i<proclist.size(); i++)
      if((n = (Node)proclist.elementAt(i)).proc != null)
        System.out.println("Load " + n.proc.getID() + ": " + n.connections);
      else
        System.out.println("Load " + n.rp.getID() + ": " + n.connections);
    return proclist;
  }


  void transfer(Node n1, Node n2, int cnt) {
    if(n1.proc != null) { // local PROC
      if(n2.proc != null)
        transfer(n1.proc, n2.proc, cnt);
      else if(n2.rp != null)
        transfer(n1.proc, n2.rp, cnt);
      else
        System.out.println("ChatMonitor error1: destination PROC null");
    }
    else if(n1.rp != null) { // remote PROC
      if(n2.proc != null)
        transfer(n1.rp, n2.proc, cnt);
      else if(n2.rp != null)
        transfer(n1.rp, n2.rp, cnt);
      else
        System.out.println("ChatMonitor error2: destination PROC null");
    }
    else
      System.out.println("ChatMonitor error: source PROC null");
  }


  void transfer(Proc proc1, Proc proc2, int cnt) {
    Proc proc;
    RemoteProc rp;
    int i;
    System.out.println("Transfer(" + cnt + ") from " + proc1.getID() + " to " +proc2.getID());
    for(i = proc1.portNo-1; (i >= 0) && (cnt > 0); i--) {
      if(proc1.port[i].connectionStatus == 1) { // connected to a local PROC
        if((proc = proc1.port[i].connectedProc).name.equals("Chat")){
          Engine.disconnectProcs(proc, 0, true);
          Engine.connectProcs(proc, 0, proc2, -1, true);
          cnt--;
        }
      }
      else if(proc1.port[i].connectionStatus==2){ // connected to a remote PROC
        if((rp = (RemoteProc)Engine.getProc(proc1.port[i].connectedProcID)).name.equals("Chat")) {
          Engine.disconnectProcs(rp, 0, true);
          Engine.connectProcs(rp, 0, proc2, -1, true);
          cnt--;
        }
      }
    }
  }

  void transfer(Proc proc1, RemoteProc rp2, int cnt) {
    Proc proc;
    RemoteProc rp;
    int i;
    System.out.println("Transfer(" + cnt + ") from " + proc1.getID() + " to " +rp2.getID());
    for(i = proc1.portNo-1; (i >= 0) && (cnt > 0); i--) {
      if(proc1.port[i].connectionStatus == 1) { // connected to a local PROC
        if((proc = proc1.port[i].connectedProc).name.equals("Chat")){
          Engine.disconnectProcs(proc, 0, true);
          Engine.connectProcs(proc, 0, rp2, -1, true);
          cnt--;
        }
      }
      else if(proc1.port[i].connectionStatus==2){ // connected to a remote PROC
        if((rp = (RemoteProc)Engine.getProc(proc1.port[i].connectedProcID)).name.equals("Chat")) {
          Engine.disconnectProcs(rp, 0, true);
          Engine.connectProcs(rp, 0, rp2, -1, true);
          cnt--;
        }
      }
    }
  }

  void transfer(RemoteProc rp1, Proc proc2, int cnt) {
    Proc proc;
    RemoteProc rp;
    int i, j;
    System.out.println("Transfer(" + cnt + ") from " + rp1.getID() + " to " +proc2.getID());
    for(i = rp1.portNo-1; (i >= 0) && (cnt > 0); i--) {
      if((j = rp1.connectedProcID[i]) != 0) {
        if((proc = Engine.getLocalProc(j)) != null) {
          // connected to a local PROC
          if(proc.name.equals("Chat")){
            Engine.disconnectProcs(proc, 0, true);
            Engine.connectProcs(proc, 0, proc2, -1, true);
            cnt--;
          }
        }
        else if((rp = Engine.getRemoteProc(j)) != null) {
          // connected to a remote PROC
          if(rp.name.equals("Chat")) {
            Engine.disconnectProcs(rp, 0, true);
            // allow time for the operation to propagate back
            try {
              Thread.currentThread().sleep(300);
            }
            catch(InterruptedException e) { }
            Engine.connectProcs(rp, 0, proc2, -1, true);
            cnt--;
          }
        }
      }
    }
  }

  void transfer(RemoteProc rp1, RemoteProc rp2, int cnt) {
    Proc proc;
    RemoteProc rp;
    int i, j;
    System.out.println("Transfer(" + cnt + ") from " + rp1.getID() + " to " +rp2.getID());
    for(i = rp1.portNo-1; (i >= 0) && (cnt > 0); i--) {
      if((j = rp1.connectedProcID[i]) != 0) {
        if((proc = Engine.getLocalProc(j)) != null) {
          // connected to a local PROC
          if(proc.name.equals("Chat")){
            Engine.disconnectProcs(proc, 0, true);
            Engine.connectProcs(proc, 0, rp2, -1, true);
            cnt--;
          }
        }
        else if((rp = Engine.getRemoteProc(j)) != null) {
          // connected to a remote PROC
          if(rp.name.equals("Chat")) {
            Engine.disconnectProcs(rp, 0, true);
            // allow time for the operation to propagate back
            try {
              Thread.currentThread().sleep(300);
            }
            catch(InterruptedException e) { }
            Engine.connectProcs(rp, 0, rp2, -1, true);
            cnt--;
          }
        }
      }
    }
  }

}


/* A Node contains a PROC or a remote PROC and the number of its connected
 * ports.
 */
class Node {
  int connections = 0;
  Proc proc = null;
  RemoteProc rp = null;

  Node(Proc p) {
    proc = p;
    for(int i=0; i<p.portNo; i++)
      if(p.port[i].connectionStatus != 0) connections++;
  }

  Node(RemoteProc p) {
    rp = p;
    for(int i=0; i<p.portNo; i++)
      if(p.connectedProcID[i] != 0) connections++;
  }
}
