/*
 * A shared whiteboard
 * @author Yizhou Cao
 */

package dacia.apps.wb;

import dacia.*;
import java.io.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.net.*;


/*
class Polyline implements Serializable { 	
  int width;
 	Color color;
  Vector points;

  Polyline(int width, Color color) {
    this.width = width;
		this.color = color;
	}        
}*/
 
class Layer implements Serializable { 
  URL   imageURL = null;
  Image image = null;
  Vector polylines;
  Vector texts;

  String name;
  String owner;
  boolean isVisible;
  boolean isReadonly;
  boolean isShared;
  Dimension minSize;

  Layer(String name, String owner) {
    this.name = name;
    this.owner = owner;
    isVisible = true;
    isReadonly = false;
    isShared = true;
	polylines = new Vector();
	texts = new Vector();
	minSize = new Dimension();
  }

  void addPolyline(Polyline pl) {
	  if (pl != null)
		  polylines.addElement(pl);
  }

  void addText(Text text) {
	  if (text != null)
		  texts.addElement(text);
  }

} 

class Polyline implements Serializable { 	
  int width;
  Color color;
  Object[] points;

  Polyline(int width, Color color) {
    this.width = width;
	this.color = color;
  }        
}

class Text implements Serializable { 	
  String str;
  int    size;
  Color  color;
  Point  base;

  Text(String str, int size, Color color, Point  base) {
    this.str = str;
    this.size = size;
	this.color = color;
	this.base = base;
  }        
}


public class Whiteboard  extends Proc {
  WhiteboardFrame frame = null;
  Hashtable layers;
  Layer currentLayer, newlayer;

  public Whiteboard() {
	  super("Whiteboard", 2);
	  frame = new WhiteboardFrame(this);
  }

  public void handleMessage(Message msg, int port) {
	  newlayer = (Layer)msg.getData();
	  newlayer.name = newlayer.owner+"."+newlayer.name;
	  OkCancelBox1 box = new OkCancelBox1(frame, "Incoming layer", 
	  		  "Accept layer "+newlayer.name+ " from "+newlayer.owner+"?");
	  if (frame.isOK) {
		  newlayer.isReadonly = true;
		  if (newlayer.imageURL != null) {
			  Toolkit toolkit = Toolkit.getDefaultToolkit();
			  newlayer.image = toolkit.getImage(newlayer.imageURL);
		  }
		  frame.panel.addLayer(newlayer);
		  frame.panel.repaint();
		  frame.controls.updateLayers();
	  }
  }

  public void handleAsyncMessage() {
	  Message msg = null;
	  while(true) {
		  if((msg = getMessage()) == null)
			  return;
		  handleMessage(msg, 0);
	  }
  }

  void sendMessage(Layer layer) {
	Image image = layer.image;
	layer.image = null;
    Message msg = new Message();
    msg.origin = getID();
    msg.setData((Object)layer);

	for (int i=0; i<portNo; i++) {
		if (port[i].connectionStatus > 0)
			output(i, msg, 1);
	}
	layer.image = image;
  }

  public void pack() {
	  layers = frame.panel.layers;
	  currentLayer = frame.panel.currentLayer;
	  frame.dispose();
	  frame = null;
  }

  public void unpack() {
	  frame = new WhiteboardFrame(this);
	  frame.panel.layers = layers;
	  frame.panel.currentLayer = currentLayer;
  }

}

class WhiteboardFrame extends Frame
    implements ActionListener {
    
  DrawPanel panel;
  DrawControls controls;
  DrawButtons buttons;
  Whiteboard  whiteboard;

  boolean  isOK;

  WhiteboardFrame(Whiteboard parent) {
		super("Whiteboard");
		whiteboard = parent;
	
		MenuBar 	menuBar;  
		Menu      menu, menu1;
		MenuItem	m;
		menuBar = new MenuBar();
		setMenuBar(menuBar);	    
	 
		// menu "File"
		menu = new Menu("File");
		menuBar.add(menu);

		m = new MenuItem("Close"); 
		m.addActionListener(this); 
		menu.add(m);

 		// menu "Draw"
		menu = new Menu("Draw");
		menuBar.add(menu);

		m = new MenuItem("Clear"); 
		m.addActionListener(this); 
		menu.add(m); 
		menu.addSeparator();

		m = new MenuItem("Layer"); 
		m.addActionListener(this); 
		menu.add(m);     

		// menu "Help"
		menu = new Menu("Help");
		m = new MenuItem("About"); 
		m.addActionListener(this); 
		menu.add(m);	
		menuBar.add(menu);
		
		// layout
		setLayout(new BorderLayout()); 		
		
		// add draw panel
		ScrollPane scrollPane = 
			new ScrollPane (ScrollPane.SCROLLBARS_AS_NEEDED);
		add("Center", scrollPane);
		panel = new DrawPanel();
		scrollPane.add(panel);
		// add control panel
      	controls = new DrawControls(panel);
		add("South",controls); 		 	
		// add botton panel
		buttons = new DrawButtons(this);
		add("East", buttons);

		// Handle window close requests
		addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				Engine.removeLocalProc(whiteboard);			
				dispose(); 
			}
		});

		setFont(new Font("sanserif", Font.PLAIN, 12));
		setSize(800, 600);
		show();
  }    	    

	public void actionPerformed(ActionEvent e) { 
		String cmd = e.getActionCommand(); 

		if (cmd.equals("Close")) {
			Engine.removeLocalProc(this.whiteboard);
			dispose();  
		}
		else if (cmd.equals("Clear"))
			panel.clear();
		else if (cmd.equals("Layer")) {
			LayerDialog layerDialog = new LayerDialog(this);
		}
		else if (cmd.equals("Send")) {
			if (panel.currentLayer.isShared)
				whiteboard.sendMessage(panel.currentLayer);
			else  {
				MessageBox box = new MessageBox((Frame)getParent(), 
								"Dacia", 
								"Cannot send "+panel.currentLayer.name+"!\n"+
								panel.currentLayer.name+" is not shared.");
			}
		}		
		else if (cmd.equals("Load")) {
			String directory = System.getProperty("user.dir");
			FileLister f = new FileLister(whiteboard, directory);
		}
		else if (cmd.equals("Unload")) {
			panel.currentLayer.image = null;
			panel.currentLayer.imageURL = null;
			panel.repaint();
		}	
		else if (cmd.equals("Redraw")) 
				panel.repaint();
		else if (cmd.equals("About"))  { 
			MessageBox about = new MessageBox((Frame)this, "About",   		
					      "Dacia--Whiteboard\n"+
						  "\n"+				  
                          "by Yizhou Cao (caoy@umich.edu)\n"+
                          "Copyright (c) 2001, U of M");		
		} 	 
	}
}

class DrawPanel extends Panel implements MouseListener, MouseMotionListener {
    public static final int LINE = 0;
    public static final int POINT = 1;
    public static final int TEXT = 2;
    public static final int NONE = 0;
    public static final int ENDPOINT = 1;     
    public static final int MIDPOINT = 2; 
	public static final int MAX_LAYER = 10;
    
    Image offscreen;
    Dimension offscreensize = new Dimension();
    Graphics offgraphics;

    int nlayer;     
    Hashtable layers = new Hashtable(10);
    Layer currentLayer;

    //Vector polylines = new Vector();
    Vector points = new Vector();
    Polyline polyline;

    int	drawMode;   
    int	snapMode; 	
    int penwidth;
    boolean firstClick = true;
    int xlast, ylast, xcurrent, ycurrent;  // last point

    public DrawPanel() {
      drawMode = LINE;   
      snapMode = NONE; 	
      penwidth = 1;
      firstClick = true;

	  setBackground(Color.white);
      setForeground(Color.black);
	  addMouseMotionListener(this);
	  addMouseListener(this);

      Layer layer0 = new Layer("layer0", Engine.getHostName());
	  layers.put("layer0", layer0);
      currentLayer = layer0;
      nlayer = 1;
    }

    void setDrawMode(int mode) {
		switch (drawMode) {
		case LINE:
		case POINT:
		case TEXT:	
			this.drawMode = mode;
			break;
		default:
			throw new IllegalArgumentException();
      }
    }

	int getDrawMode() {
		return drawMode;
	}

    void setPenwidth(int width) {
      penwidth = width;
    }

    int getPenwidth() {
      return penwidth;
    }

    void setImage(URL url) {
		Toolkit toolkit = Toolkit.getDefaultToolkit();
		currentLayer.image = toolkit.getImage(url);
		currentLayer.imageURL = url;
		repaint();
	}

    void setCurrentLayer(String lname) {
      currentLayer = getLayer(lname);
    }

    void addLayer(Layer layer) {
      layers.put(layer.name, layer);
	  nlayer++;
    }

    void removeLayer(String lname) {
      layers.remove(lname);
	  nlayer--;
    }

    Layer getLayer(String lname) {
		Layer layer;
		layer = (Layer)layers.get(lname);
		return layer;
	}

    void clear() {
      currentLayer.polylines.clear();
      points.clear();
      repaint();
    }

    public void mouseClicked(MouseEvent e) { 
		if (drawMode == TEXT) {
			Point base = new Point(e.getX(), e.getY());
			TextInputDialog dlg = 
				new TextInputDialog((Frame)(getParent().getParent()), this, base);
		}
    } 	

    public void mouseDragged(MouseEvent e) {
        e.consume();
        if (e.getModifiers()==MouseEvent.BUTTON1_MASK) {
          switch (drawMode) {
            case LINE:
            case TEXT:
              break;
            case POINT:
            default:
				int x = e.getX();
				int y = e.getY();
                points.addElement(new Point(x, y));
				currentLayer.minSize.width =Math.max(currentLayer.minSize.width , x);
 				currentLayer.minSize.height=Math.max(currentLayer.minSize.height, y);
				repaint();
                break;
          }
        }      
    }

    public void mouseMoved(MouseEvent e) {
        e.consume();
        if (drawMode == LINE && !firstClick) {
            xcurrent = e.getX();
            ycurrent = e.getY();
            repaint();
        }
    }

    public void mousePressed(MouseEvent e) {
        e.consume();
        if (e.getModifiers()==MouseEvent.BUTTON1_MASK) {
          switch (drawMode) {
            case LINE:
		    case TEXT:
                break;
            case POINT:
            default:
				int x = e.getX();
				int y = e.getY();
                polyline = new Polyline(penwidth, getForeground());
                points.clear(); 
                points.addElement(new Point(x, y));
				currentLayer.minSize.width =Math.max(currentLayer.minSize.width , x);
 				currentLayer.minSize.height=Math.max(currentLayer.minSize.height, y);
                repaint();
                break;
          }
        }
    }

    public void mouseReleased(MouseEvent e) {
        e.consume(); 		
        switch (e.getModifiers()) {
        case MouseEvent.BUTTON1_MASK: 	
            switch (drawMode) {
            case LINE: 		
                if (firstClick) {
                    polyline = new Polyline(penwidth, getForeground());
                    firstClick = false;
                }
                xlast = xcurrent = e.getX();
                ylast = ycurrent = e.getY();
                points.addElement(new Point(xlast, ylast));
				currentLayer.minSize.width =
					Math.max(currentLayer.minSize.width , xlast);
 				currentLayer.minSize.height=
					Math.max(currentLayer.minSize.height, ylast);
                repaint();
                break;
            case POINT:
                points.addElement(new Point(e.getX(), e.getY()));
				currentLayer.minSize.width =
					Math.max(currentLayer.minSize.width , e.getX());
 				currentLayer.minSize.height=
					Math.max(currentLayer.minSize.height, e.getY());
                polyline.points = points.toArray();
                //polyline.points = points;
                currentLayer.polylines.addElement(polyline);
                points.clear();
                repaint(); 			
                break; 
            case TEXT:
            default:
                break; 
            } 
            break;
        case MouseEvent.BUTTON3_MASK:
            if (drawMode == LINE && !firstClick) {
                firstClick = true;
                if (points.size()>1) {
                  polyline.points = points.toArray();
                  //polyline.points = points;
                  currentLayer.polylines.addElement(polyline);
                }
                points.clear();
                repaint();
            }
            break;
        default:
            break;
        }
    }

    public void mouseEntered(MouseEvent e) {
    }

    public void mouseExited(MouseEvent e) {
    }

    private Dimension getMinSize() {
		Layer layer;
		Dimension d = new Dimension();
		for (Enumeration layerEnum=layers.elements(); 
			 layerEnum.hasMoreElements(); ) {
			layer = (Layer)layerEnum.nextElement();
			if (layer.isVisible) {
			  d.width = Math.max(d.width,  layer.minSize.width);
			  d.height= Math.max(d.height, layer.minSize.height);
			  if (layer.image != null) {
				  d.width = Math.max(d.width, layer.image.getWidth(this) +10);
				  d.height = Math.max(d.height, layer.image.getHeight(this) +10);
			  }
			}
		}
		return d;
	}

    private void drawPoint(Graphics g, int x, int y, int width) {
       if (width > 1) {
           int w = (width+1)/2;
           g.fillOval(x-w, y-w, 2*w+1, 2*w+1);
       }
       else
           g.fillOval(x, y, 1, 1);
    }

    private void drawLine(Graphics g, int x1, int y1, int x2, int y2, int width) {
      int i;
      g.drawLine(x1, y1, x2, y2);
      // horizontal line
      if (y2-y1 == 0) {
        for (i=1; i<width; i+=1) {
          g.drawLine(x1-i, y1, x2-i, y2);
          g.drawLine(x1+i, y1, x2+i, y2);
        }
      }
      // vertical line
      else if (x2-x1 == 0) {
        for (i=1; i<width; i+=1) {
          g.drawLine(x1, y1-i, x2, y2-i);
          g.drawLine(x1, y1+i, x2, y2+i);
        }
      }      
     
      if ((y2-y1)*(x2-x1) > 0) { 
        for (i=1; i<=width/2; i+=1) {
          g.drawLine(x1-i+1, y1+i,   x2-i+1, y2+i);
          g.drawLine(x1+i,   y1-i+1, x2+i,   y2-i+1);
          if (i*2 < width) { 
            g.drawLine(x1-i, y1+i, x2-i, y2+i);
            g.drawLine(x1+i, y1-i, x2+i, y2-i);
          }
        }
      }
      else {
        for (i=1; i<=width/2; i+=1) {
          g.drawLine(x1-i+1, y1-i,   x2-i+1, y2-i);
          g.drawLine(x1+i,   y1+i-1, x2+i,   y2+i-1);
          if (i*2 < width) { 
            g.drawLine(x1+i, y1+i, x2+i, y2+i);
            g.drawLine(x1-i, y1-i, x2-i, y2-i);
          }
        }
      } 

      if (width > 1) {
        drawPoint(g, x1, y1, width);
        drawPoint(g, x2, y2, width);
      }
    } 

	private void drawPolyline(Graphics g, Polyline pl) {
		int npoint;
		int x1, y1, x2, y2;

		g.setColor(pl.color);
		//npoint = pl.points.size();
		npoint = pl.points.length;
		if (npoint > 0) {
			//x1 = ((Point)pl.points.elementAt(0)).x;
            //y1 = ((Point)pl.points.elementAt(0)).y;
            x1 = ((Point)pl.points[0]).x;
            y1 = ((Point)pl.points[0]).y;
            drawPoint(g, x1, y1, pl.width);
            for (int i=1; i<npoint; i++) {
				// draw each line segment
				//x2 = ((Point)pl.points.elementAt(j)).x;
				//y2 = ((Point)pl.points.elementAt(j)).y;
				x2 = ((Point)pl.points[i]).x;
				y2 = ((Point)pl.points[i]).y;
				drawLine(g, x1, y1, x2, y2, pl.width);
				x1 = x2;
				y1 = y2;
			}
		}
	}

	private void drawText(Graphics g, Text text) {
		g.setColor(text.color);
		g.setFont(new Font("sanserif", Font.PLAIN, text.size));
		g.drawString(text.str, text.base.x, text.base.y);
	}

    public void paint(Graphics g) { 
        update(g);
    }
    
    public synchronized void update(Graphics g) { 
      Dimension panelSize = getSize();
	  Dimension minSize = getMinSize();
	  Dimension	vpSize = ((ScrollPane)getParent()).getViewportSize();

      if ((offscreen == null) 
		  || (panelSize.width  != offscreensize.width) 
		  || (panelSize.height != offscreensize.height) 
		  || (offscreensize.width  < minSize.width)
		  || (offscreensize.height < minSize.height) ) {
		  offscreensize.width  = Math.max(vpSize.width, minSize.width);
		  offscreensize.height = Math.max(vpSize.height, minSize.height);
	      offscreen = createImage(offscreensize.width+10, offscreensize.height+10);
        if (offgraphics != null) {
	        offgraphics.dispose();
        }
        offgraphics = offscreen.getGraphics();
        //offgraphics.setFont(getFont());
      }

      offgraphics.setColor(Color.white);
      offgraphics.fillRect(0, 0, offscreensize.width, offscreensize.height);

      Layer layer;
      Vector pls, texts;
      Polyline pl;
	  Text text;
      int i, j;
      int npoint, npl, ntext;
      int x1, y1, x2, y2;

      for (Enumeration layerEnum=layers.elements();layerEnum.hasMoreElements();) {
		  // for each layer
		  layer = (Layer)layerEnum.nextElement();

		  if (layer.isVisible) { 
			// draw the image
			if (layer.image != null)
				offgraphics.drawImage(layer.image, 10, 10, this);      

			// draw polylines 
			pls = layer.polylines;
			npl = pls.size();
			if (npl > 0)
              for (i=0; i<npl; i++) { 
				pl = (Polyline)pls.elementAt(i);
				drawPolyline(offgraphics, pl);
			  } 
			// draw texts 
			texts = layer.texts;
			ntext = texts.size();
			if (ntext > 0)
              for (i=0; i<ntext; i++) { 
				text = (Text)texts.elementAt(i);
				drawText(offgraphics, text);
			  }
		  }
      }
      offgraphics.setColor(getForeground());
      npoint = points.size();
      if (npoint > 0) {
        x1 = ((Point)points.elementAt(0)).x;
        y1 = ((Point)points.elementAt(0)).y;
        drawPoint(offgraphics, x1, y1, penwidth);
        for (i=1; i<npoint; i++) {
          x2 = ((Point)points.elementAt(i)).x;
          y2 = ((Point)points.elementAt(i)).y;
          drawLine(offgraphics, x1, y1, x2, y2, penwidth);
          x1 = x2;
          y1 = y2;
        }
      }         
        
      if (drawMode == LINE && !firstClick) {
        offgraphics.setColor(getForeground());
        drawLine(offgraphics, xlast, ylast, xcurrent, ycurrent, penwidth);
      }
      
	  setSize(offscreensize);
      g.drawImage(offscreen, 0, 0, null);
    }
}


class DrawControls extends Panel implements ItemListener {
  DrawPanel target;
  Choice layers;
  Choice shapes;
  Choice penwidths;
  Choice colors;

  public DrawControls(DrawPanel target) {
	  Label l;
	  Layer layer;
      this.target = target;
      setLayout(new FlowLayout());
      setBackground(Color.lightGray);

      l = new Label("Current layer:");
      add(l);
      layers = new Choice();
      layers.addItemListener(this);
      for (Enumeration layerEnum=target.layers.elements(); 
           layerEnum.hasMoreElements(); ) {
		  layer = (Layer)layerEnum.nextElement();
		  if (!layer.isReadonly && layer.isVisible)
			  layers.addItem(layer.name);
      }
	  layers.select(target.currentLayer.name);
      add(layers);
      
      l = new Label("  Shape:");
      add(l);
      shapes = new Choice();
      shapes.addItemListener(this);
      shapes.addItem("Line");
      shapes.addItem("Point");
      shapes.addItem("Text");
      shapes.setBackground(Color.lightGray);
      add(shapes);

      l = new Label("  Pen size:");
      add(l);
      penwidths = new Choice();
      penwidths.addItemListener(this);
      for (int i=1; i<20; i++) 
        penwidths.addItem(String.valueOf(i));
      penwidths.setBackground(Color.lightGray);
      add(penwidths);

      l = new Label("  Color:");
      add(l);
      colors = new Choice();
      colors.addItemListener(this);
      colors.addItem("black");
      colors.addItem("blue");
      colors.addItem("gray");
      colors.addItem("green");
      colors.addItem("magenta");
      colors.addItem("orange");
      colors.addItem("pink");
      colors.addItem("red");
      colors.addItem("yellow");
      colors.setBackground(Color.lightGray);
      add(colors);
  }
    
  public void updateLayers() {
	  Layer layer;
      layers.removeAll();
      for (Enumeration layerEnum=target.layers.elements(); 
           layerEnum.hasMoreElements(); ) {
		  layer = (Layer)layerEnum.nextElement();
		  if (!layer.isReadonly && layer.isVisible)
			  layers.addItem(layer.name);
      }
	  layers.select(target.currentLayer.name);
	  repaint();
  }

  public void paint(Graphics g) {
      Rectangle r = getBounds();
      g.setColor(Color.lightGray);
      g.draw3DRect(0, 0, r.width, r.height, true);
  }
    
  public void itemStateChanged(ItemEvent e) {
    if (e.getSource() == layers)
      target.setCurrentLayer(layers.getSelectedItem());
    else if (e.getSource() == penwidths) {
	  int w = Integer.parseInt(penwidths.getSelectedItem());
      target.setPenwidth(w);
	  if (target.getDrawMode() == DrawPanel.TEXT)
		  target.setFont(new Font("sanserif", Font.PLAIN, w));
	}
    else if (e.getSource() == colors) {
      String colorName = colors.getSelectedItem();
      if (colorName.equals("black"))
        target.setForeground(Color.black);
      else if (colorName.equals("blue"))
        target.setForeground(Color.blue);
      else if (colorName.equals("gray"))
        target.setForeground(Color.gray);
      else if (colorName.equals("green"))
        target.setForeground(Color.green);
      else if (colorName.equals("magenta"))
        target.setForeground(Color.magenta);
      else if (colorName.equals("orange"))
        target.setForeground(Color.orange);
      else if (colorName.equals("pink"))
        target.setForeground(Color.pink);
      else if (colorName.equals("red"))
        target.setForeground(Color.red);
      else if (colorName.equals("yellow"))
        target.setForeground(Color.yellow);
    }
    else if (e.getSource() == shapes) {
      String choice = (String) e.getItem();
      if (choice.equals("Line")) {
		  if (target.getDrawMode() == DrawPanel.TEXT) {
			  target.setPenwidth(1);
			  penwidths.select(1);
		  }
          target.setDrawMode(DrawPanel.LINE);
	  }
      else if (choice.equals("Point")) {
		  if (target.getDrawMode() == DrawPanel.TEXT) {
			  target.setPenwidth(1);
			  penwidths.select(1);
		  } 
          target.setDrawMode(DrawPanel.POINT);
	  }
      else if (choice.equals("Text")) {
          target.setDrawMode(DrawPanel.TEXT);
		  target.setPenwidth(12);
		  penwidths.select("12");
		  target.setFont(new Font("sanserif", Font.PLAIN, 12));
	  }
    }
  }
}

class DrawButtons extends Panel {

  public DrawButtons(WhiteboardFrame frame) {
	  GridBagLayout g = new GridBagLayout();
	  setLayout(g);
	  GridBagConstraints gbc = new GridBagConstraints(); 
      setBackground(Color.lightGray);

	  Button b;
	  b = new Button("Layer");
	  b.addActionListener(frame); 
	  gbc.insets = new Insets(8, 5, 0, 5);
	  gbc.gridx = 1; 	
	  gbc.gridy = 1;
	  gbc.gridheight = 1; 
	  gbc.gridwidth = 1;  	
	  g.setConstraints(b, gbc); 
	  add(b);

	  b = new Button("Send");
	  b.addActionListener(frame); 
	  gbc.gridy = 2;
	  g.setConstraints(b, gbc); 
	  add(b);

	  b = new Button("Load");
	  b.addActionListener(frame); 
	  gbc.gridy = 3;
	  g.setConstraints(b, gbc); 
	  add(b);

	  b = new Button("Unload");
	  b.addActionListener(frame); 
	  gbc.gridy = 4;
	  g.setConstraints(b, gbc); 
	  add(b);

	  b = new Button("Pan");
	  gbc.insets = new Insets(40, 5, 0, 5);
	  b.addActionListener(frame); 
	  gbc.gridy = 5;
	  g.setConstraints(b, gbc); 
	  add(b);

	  b = new Button("Zoom");
	  gbc.insets = new Insets(8, 5, 0, 5);
	  b.addActionListener(frame); 
	  gbc.gridy = 6;
	  g.setConstraints(b, gbc); 
	  add(b);

  }
    
  public void paint(Graphics g) {
      Rectangle r = getBounds();
      g.setColor(Color.lightGray);
      g.draw3DRect(0, 0, r.width, r.height, true);
  }
    
}


class LayerDialog extends Dialog 
	implements ActionListener, ItemListener { 

  DrawPanel targetPanel;
  DrawControls targetControls;
  Panel layerPanel;   
  Label  current;
  CheckboxGroup cbg = new CheckboxGroup();
  Checkbox[] layerBox; 
  Checkbox[] visibleBox; 
  Checkbox[] readonlyBox; 
  Checkbox[] shareBox; 
  Label[] ownerLabel;
  int  nlayer=0;  
 
  public LayerDialog(Frame parent) {
    super(parent, "Layer", true); 
    targetPanel = ((WhiteboardFrame)parent).panel;
    targetControls = ((WhiteboardFrame)parent).controls;
	layerBox = new Checkbox[targetPanel.MAX_LAYER]; 
	visibleBox = new Checkbox[targetPanel.MAX_LAYER]; 
    readonlyBox = new Checkbox[targetPanel.MAX_LAYER]; 
    shareBox = new Checkbox[targetPanel.MAX_LAYER]; 
    ownerLabel = new Label[targetPanel.MAX_LAYER];  
	Label l;
    Button b;

    setFont(new Font("sanserif", Font.PLAIN, 12));
    setBackground(Color.lightGray);

    GridBagLayout g = new GridBagLayout();
    setLayout(g);
    GridBagConstraints gbc = new GridBagConstraints(); 

	l = new Label("Layer");
	gbc.insets = new Insets(10, 30, 0, 30);
    gbc.gridx = 1; 	
    gbc.gridy = 1;
    gbc.gridheight = 1; 
    gbc.gridwidth = 1;  	
    gbc.anchor = GridBagConstraints.NORTH;
    g.setConstraints(l, gbc); 
	add(l);

	l = new Label("Owner");
	gbc.insets = new Insets(10, 0, 0, 0);
    gbc.gridx = 2; 		
    g.setConstraints(l, gbc); 
	add(l);

	l = new Label("On");
    gbc.gridx = 3; 		
    g.setConstraints(l, gbc); 
	add(l);
	
	l = new Label("Share");
    gbc.gridx = 4; 		
    g.setConstraints(l, gbc); 
	add(l);
	
	l = new Label("Read-only");
    gbc.gridx = 5; 		
    g.setConstraints(l, gbc); 
	add(l);
	
	//layerPanel = new ScrollPane();
	ScrollPane scroll = new ScrollPane();
    layerPanel = new Panel();
	scroll.add(layerPanel);
	gbc.insets = new Insets(0, 10, 0, 10);
	gbc.weightx = 1; 
    gbc.weighty = 1; 	
    gbc.gridx = 1; 	
    gbc.gridy = 2;
    gbc.gridwidth = 5;
    gbc.anchor = GridBagConstraints.NORTH;
  	gbc.fill = GridBagConstraints.BOTH;
    g.setConstraints(scroll, gbc); 
	add(scroll);
	
	current = new Label("Current layer: "+targetPanel.currentLayer.name); 
	gbc.insets = new Insets(10, 10, 0, 0);
    gbc.gridx = 1; 	
    gbc.gridy = 3;
    gbc.gridheight = 1; 
    gbc.gridwidth = 5;  	
	gbc.weightx = 0; 
    gbc.weighty = 0; 	
    gbc.anchor = GridBagConstraints.WEST;
  	gbc.fill = GridBagConstraints.NONE;
    g.setConstraints(current, gbc); 
	add(current);

    b = new Button("Current");  	
    b.addActionListener(this);
	gbc.insets = new Insets(10, 10, 0, 0);
    gbc.gridx = 1; 	
    gbc.gridy = 4;
    gbc.gridheight = 1; 
    gbc.gridwidth = 1;  	
    g.setConstraints(b, gbc); 
    add(b);  		

    b = new Button("Delete");  	
    b.addActionListener(this);
    gbc.gridx = 2; 	
    gbc.gridy = 4;
    g.setConstraints(b, gbc); 
    add(b);  		

    b = new Button("Modify");  	
    b.addActionListener(this);
    gbc.gridx = 3; 	
    gbc.gridy = 4;
    g.setConstraints(b, gbc); 
    add(b);  		

    b = new Button("New");  	
    b.addActionListener(this);
    gbc.gridx = 4; 	
    gbc.gridy = 4;
    gbc.gridwidth = 2;  	
    g.setConstraints(b, gbc); 
    add(b);  		

    b = new Button("OK");  	
    b.addActionListener(this);
	gbc.insets = new Insets(10, 10, 10, 0);
    gbc.gridx = 2; 	
    gbc.gridy = 5;
    gbc.gridwidth = 1;  	
    g.setConstraints(b, gbc); 
    add(b);  		

    b = new Button("Cancel");  	
    b.addActionListener(this);	 	
    gbc.gridx = 3;
    g.setConstraints(b, gbc); 
    add(b);   	

	addLayers();
	//scroll.setSize(scroll.getMaximumSize());

	pack();
	//setResizable(false); 
	show();
  } 

  private void addLayers() {	
	Layer layer;
    Enumeration layerEnum;
	int i;

    GridBagLayout g1 = new GridBagLayout();
    layerPanel.setLayout(g1);
    GridBagConstraints gbc1 = new GridBagConstraints();
	Label l;

	l = new Label("---------------------");
    gbc1.gridx = 1; 	
    gbc1.gridy = 1;
    gbc1.gridheight = 1; 
    gbc1.gridwidth = 1;  	
    gbc1.anchor = GridBagConstraints.WEST;
  	gbc1.fill = GridBagConstraints.BOTH;
    g1.setConstraints(l, gbc1); 
	layerPanel.add(l);

	l = new Label("------------");
    gbc1.gridx = 2; 		
    g1.setConstraints(l, gbc1); 
	layerPanel.add(l);

	l = new Label("-------");
    gbc1.gridx = 3; 		
    g1.setConstraints(l, gbc1); 
	layerPanel.add(l);
	
	l = new Label("----------");
    gbc1.gridx = 4; 		
    g1.setConstraints(l, gbc1); 
	layerPanel.add(l);
	
	l = new Label("----------");
    gbc1.gridx = 5; 		
    g1.setConstraints(l, gbc1); 
	layerPanel.add(l);

	for (layerEnum =targetPanel.layers.elements(), i=0, nlayer=0; 
         layerEnum.hasMoreElements() && nlayer<targetPanel.MAX_LAYER; 
		 i++ ) {
        layer = (Layer)layerEnum.nextElement();
		nlayer++;

		layerBox[i] = new Checkbox(layer.name, cbg, false);
		gbc1.gridx = 1;
		gbc1.gridy = i+2;
		gbc1.anchor = GridBagConstraints.WEST;
		gbc1.fill = GridBagConstraints.NONE;
		g1.setConstraints(layerBox[i], gbc1);
		layerPanel.add(layerBox[i]);

		ownerLabel[i] = new Label(layer.owner);
		gbc1.gridx = 2;
		g1.setConstraints(ownerLabel[i], gbc1);
		layerPanel.add(ownerLabel[i]);

		visibleBox[i] = new Checkbox("", null, layer.isVisible);
		visibleBox[i].addItemListener(this);	 	
		gbc1.gridx = 3;
		gbc1.anchor = GridBagConstraints.CENTER;
		g1.setConstraints(visibleBox[i], gbc1);
		layerPanel.add(visibleBox[i]);

		shareBox[i] = new Checkbox("", null, layer.isShared);
		gbc1.gridx = 4;
		g1.setConstraints(shareBox[i], gbc1);
		layerPanel.add(shareBox[i]);
		
		readonlyBox[i] = new Checkbox("", null, layer.isReadonly);
		readonlyBox[i].addItemListener(this);	 	
		gbc1.gridx = 5;
		g1.setConstraints(readonlyBox[i], gbc1);
		layerPanel.add(readonlyBox[i]);
	}

	l = new Label("        ");
    gbc1.gridx = 1;
    gbc1.gridy = 1; 		
    g1.setConstraints(l, gbc1); 
	layerPanel.add(l);
	//layerPanel.getPreferredSize();
  }

  public void updateLayers() {
	  layerPanel.removeAll();
	  addLayers();
	  repaint();
  }

  public void update(Graphics g) {
	  paintAll(g);
  }

  public void actionPerformed(ActionEvent e) {
	Layer layer = null;
	String lname = null;
	String label = ((Button)e.getSource()).getLabel(); 
 
	if (cbg.getSelectedCheckbox()!=null) {
		lname = cbg.getSelectedCheckbox().getLabel();
		layer = targetPanel.getLayer(lname);
	}

    if (label.equals("Current") && lname!=null) {
		if (targetPanel.getLayer(lname).isReadonly) {
			MessageBox box = new MessageBox((Frame)getParent(), 
								"Dacia", 
								"Cannot change "+lname+" to current layer!\n"+
								lname+" is read-only.");
		}
        targetPanel.setCurrentLayer(lname);
		current.setText("Current layer: "+lname);
	}
    if (label.equals("Delete") && layer!=null) {
		if (lname.equals(targetPanel.currentLayer.name)) {
			MessageBox box = new MessageBox((Frame)getParent(), 
										 "Dacia", 
										 "Cannot delect "+lname+"!\n"+
											lname+" is the current layer.");
		}	
		else {
			nlayer--;
			targetPanel.removeLayer(lname);
			updateLayers();
		}
    }
    if (label.equals("Modify") && layer!=null) {
		if (layer.isReadonly) {
			MessageBox box = new MessageBox((Frame)getParent(), 
										 "Dacia", 
										 "Cannot modify "+lname+"!\n"+
											lname+" is read-only.");
		}	
		else {
			ChangeLayerDialog dlg = 
				new ChangeLayerDialog(this, targetPanel, lname);
			updateLayers();
		}
	}
    if (label.equals("New")) {
		ChangeLayerDialog dlg = 
			new ChangeLayerDialog(this, targetPanel, null);
		updateLayers();
	}
	if (label.equals("OK")) {
		for (int i=0; i<nlayer; i++) {
			lname = layerBox[i].getLabel();
			layer = targetPanel.getLayer(lname);
			if (layer != null) {
				layer.isVisible = visibleBox[i].getState();
				layer.isReadonly = readonlyBox[i].getState();
				layer.isShared = shareBox[i].getState();
			}
		}
		dispose(); 
		targetControls.updateLayers();
		targetPanel.repaint();
	}
	if (label.equals("Cancel")) 
		dispose(); 
  }

  public void itemStateChanged(ItemEvent e) {
	int    i;
	String lname, ownername;
	Layer  layer;

	Object src = e.getSource();
	for (i=0; i<nlayer; i++) {
		if (src==visibleBox[i]) {
			lname = layerBox[i].getLabel();
			if (lname.equals(targetPanel.currentLayer.name)) {
				MessageBox box = new MessageBox((Frame)getParent(), 
										 "Dacia", 
										 "Cannot turn off "+lname+"!\n"+
											lname+" is the current layer.");
				visibleBox[i].setState(true);
			}
		}
		if (src==readonlyBox[i]) {
			lname = layerBox[i].getLabel();
			ownername = ownerLabel[i].getText();
			if (lname.equals(targetPanel.currentLayer.name)) {
				MessageBox box = new MessageBox((Frame)getParent(), 
										 "Dacia", 
										 "Cannot make "+lname+" read-only!\n"+
											lname+" is the current layer.");
				readonlyBox[i].setState(false);
			}
			if (!ownername.equals(Engine.getHostName())) {
				MessageBox box = new MessageBox((Frame)getParent(), 
										 "Dacia", 
										 "Cannot write on layer "+lname+"!\n"+
											lname+" is owned by others.");
				readonlyBox[i].setState(true);
			}
		}
	}
  }
				
} 


class ChangeLayerDialog extends Dialog implements ActionListener {
    
	DrawPanel   target;
	String      oldname;

    TextField   layername; 	
    Checkbox    current;

    public ChangeLayerDialog(Dialog parent, DrawPanel target, String oldname) { 
		super(parent, "New Layer", true); 
		this.target = target;
		this.oldname = oldname;
		setFont(new Font("sanserif", Font.PLAIN, 12));
		setBackground(Color.lightGray);
		
		GridBagLayout g = new GridBagLayout();
		setLayout(g);	
		GridBagConstraints gbc = new GridBagConstraints(); 
		
		Label l; 
		l = new Label("Name"); 
		gbc.insets = new Insets(10, 10, 0, 0);
		gbc.weightx = 1.0; 
		gbc.weighty = 1.0; 	
		gbc.gridx = 1; 	
		gbc.gridy = 1;
		gbc.gridheight = 1; 
		gbc.gridwidth = 1;  	
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		layername = new TextField(12); 		
		gbc.insets = new Insets(10, 10, 0, 10);
		gbc.gridx = 2; 		
		g.setConstraints(layername, gbc); 
		add(layername); 
		if (oldname!=null)
			layername.setText(oldname);
				
		current = new Checkbox("Set as current layer", true);
		gbc.insets = new Insets(0, 10, 0, 0);
		gbc.gridx = 1; 		
		gbc.gridy = 2;	
		gbc.gridwidth = 2;  	
		g.setConstraints(current, gbc); 
		add(current); 
				
		Button ok = new Button("OK");  	
		ok.addActionListener(this);
		gbc.insets = new Insets(10, 10, 10, 0);
		gbc.gridx = 1; 	
		gbc.gridy = 4;
		gbc.gridwidth = 1;  	
		g.setConstraints(ok, gbc); 
		add(ok);  		
		
		Button cancel = new Button("Cancel");  	
		cancel.addActionListener(this);	 	
		gbc.gridx = 2;
		g.setConstraints(cancel, gbc); 
		add(cancel);   	
		
		pack();
		setResizable(false); 
		show();
    } 

    public void actionPerformed(ActionEvent e) {
		String label = ((Button)e.getSource()).getLabel(); 
		Layer  layer;
		String newname = layername.getText();
		if (label.equals("OK") && !newname.equals("")) {
			if (target.getLayer(newname) != null) {
				MessageBox box = new MessageBox((Dialog)getParent(), 
										 "Dacia", 
										 "Layer "+newname+" exists. \n"+
											"Please choose another name.");
			}
			else {
			  if (oldname != null) { 	// if rename
				layer = (Layer)target.getLayer(oldname);
				target.removeLayer(oldname);
				layer.name = newname;
		  	  }
			  else {
				layer = new Layer(newname, Engine.getHostName());
				((LayerDialog)getParent()).nlayer++;
			  }

			  target.addLayer(layer);
			  if (current.getState())
				target.setCurrentLayer(newname);
			  dispose();
			}
		}
		else if  (label.equals("OK"))
			dispose();
	}
}


class TextInputDialog extends Dialog implements ActionListener {
    
	DrawPanel  target;
    TextArea   input; 	
	Point base;

    public TextInputDialog(Frame parent, DrawPanel target, Point base) { 
		super(parent, "Input text", true); 
		this.target = target;
		this.base = base;
		setFont(new Font("sanserif", Font.PLAIN, 12));
		setBackground(Color.lightGray);
		
		GridBagLayout g = new GridBagLayout();
		setLayout(g);	
		GridBagConstraints gbc = new GridBagConstraints(); 
		
		Label l; 
		l = new Label("Input your text:"); 
		gbc.insets = new Insets(10, 10, 0, 0);
		gbc.weightx = 0.0; 
		gbc.weighty = 0.0; 	
		gbc.gridx = 1; 	
		gbc.gridy = 1;
		gbc.gridheight = 1; 
		gbc.gridwidth = 2;  	
		gbc.anchor = GridBagConstraints.WEST;
		g.setConstraints(l, gbc); 
		add(l); 
	
		input = new TextArea(6, 30); 		
		gbc.insets = new Insets(0, 10, 0, 10);
		gbc.weightx = 1.0; 
		gbc.weighty = 1.0; 	
		gbc.gridy = 2; 		
		g.setConstraints(input, gbc); 
		add(input); 
								
		Button ok = new Button("OK");  	
		ok.addActionListener(this);
		gbc.insets = new Insets(10, 10, 10, 0);
		gbc.weightx = 0.0; 
		gbc.weighty = 0.0; 	
		gbc.gridx = 1; 	
		gbc.gridy = 3;
		gbc.gridwidth = 1;  	
		gbc.anchor = GridBagConstraints.CENTER;
		g.setConstraints(ok, gbc); 
		add(ok);  		
		
		Button cancel = new Button("Cancel");  	
		cancel.addActionListener(this);	 	
		gbc.gridx = 2;
		g.setConstraints(cancel, gbc); 
		add(cancel);   	
		
		pack();
		show();
    } 

    public void actionPerformed(ActionEvent e) {
		String label = ((Button)e.getSource()).getLabel(); 
		String str = input.getText();
		if (label.equals("OK") && !str.equals("")) {
			Text text = new Text(str, target.getPenwidth(),
								 target.getForeground(), base);
			target.currentLayer.addText(text);						 
		}
		dispose();
		target.repaint();
	}
}

class OkCancelBox1 extends Dialog implements ActionListener {
 	   	 
    String msg;
    WhiteboardFrame frame;  	 
    
    public OkCancelBox1(Frame parent, String title, String msg) { 
		super(parent, title, true);  
		frame = (WhiteboardFrame)parent;	  
		this.msg = msg;
		setFont(new Font("sanserif", Font.PLAIN, 12));
		
		setLayout(new BorderLayout(15, 15));	
		add("Center", new MultiLineLabel(msg, 20, 20));

		Panel p = new Panel();  	
		Button ok = new Button("OK"); 	
		ok.addActionListener(this);
		p.add(ok);
		Button cancel = new Button("Cancel"); 	 
		cancel.addActionListener(this);
		p.add(cancel);
		add("South", p); 
		
		setSize(240, 120);
		setResizable(false); 
		show(); 	 
    }
 	
    public void paint(Graphics g) { 	
		g.drawString(msg, 10, 60);	 
    }

    public void actionPerformed(ActionEvent e) {
		String label = ((Button)e.getSource()).getLabel(); 		
		if (label.equals("OK")) { 	
			frame.isOK = true;
			dispose(); 
		}	
		if (label.equals("Cancel")) {
			frame.isOK = false;
			dispose(); 
		}
    } 	
}    

